#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Class for window selection on asdf file and handles parallel I/O
so they are invisible to users.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""

from __future__ import (absolute_import, division, print_function)
from functools import partial
import os
import inspect
from copy import deepcopy
import json
import pyflex
from ..window.window import window_on_stream
from ..window.utils import merge_windows, stats_all_windows
from ..window.io import get_json_content, WindowEncoder
from .utils import smart_mkdir
from .procbase import ProcASDFBase


def check_param_keywords(config):
    deletes = ["self", "noise_start_index", "noise_end_index",
               "signal_start_index", "signal_end_index",
               "window_weight_fct"]

    default_keywords = inspect.getfullargspec(pyflex.Config.__init__).args
    for d in deletes:
        default_keywords.remove(d)

    print("DK: ", set(default_keywords))
    print("CK: ", set(config.keys()))

    if set(default_keywords) != set(config.keys()):
        print("Missing: %s" % (set(default_keywords) - set(config.keys())))
        print("Redundant: %s" % (set(config.keys()) - set(default_keywords)))

        if len(set(default_keywords) - set(config.keys())) > 0:
            raise ValueError("config file is missing values compared to "
                             "pyflex.Config")


def load_window_config(param):
    config_dict = {}
    flag_list = []

    for key, value in param.items():
        # pop the "instrument_merge_flag" value out
        flag_list.append(value["instrument_merge_flag"])
        value.pop("instrument_merge_flag")

        check_param_keywords(value)
        config_dict[key] = pyflex.Config(**value)

    if not all(_e == flag_list[0] for _e in flag_list):
        raise ValueError("Instrument_merge_flag not consistent amonge"
                         "different parameter yaml files(%s). Check!"
                         % flag_list)

    return config_dict, flag_list[0]


def write_window_json(results, output_file):

    print("Output window file: %s" % output_file)
    window_all = {}
    for station, sta_win in results.items():
        if sta_win is None:
            continue
        window_all[station] = {}
        _window_comp = {}
        for trace_id, trace_win in sta_win.items():
            _window = [get_json_content(_i, simple_mode=False)
                       for _i in trace_win]
            _window_comp[trace_id] = _window
        window_all[station] = _window_comp

    with open(output_file, 'w') as fh:
        j = json.dumps(window_all, cls=WindowEncoder, sort_keys=True,
                       indent=2, separators=(',', ':'))
        try:
            fh.write(j)
        except TypeError:
            fh.write(j.encode())


def window_wrapper(obsd_station_group, synt_station_group, config_dict=None,
                   obsd_tag=None, synt_tag=None, user_modules=None,
                   event=None, figure_mode=False, figure_dir=None,
                   _verbose=False):
    """
    Wrapper for asdf I/O
    """

    # Make sure everything thats required is there.
    try:
        hasattr(synt_station_group, "StationXML")
    except Exception:
        print("Missing StationXML from synt_staiton_group")
        return
    try:
        hasattr(obsd_station_group, obsd_tag)
    except Exception:
        print("Missing tag '%s' from obsd_station_group" % obsd_tag)
        return
    try:
        hasattr(synt_station_group, synt_tag)
    except Exception:
        print("Missing tag '%s' from synt_station_group" % synt_tag)
        return

    inv = synt_station_group.StationXML
    observed = getattr(obsd_station_group, obsd_tag)
    synthetic = getattr(synt_station_group, synt_tag)

    return window_on_stream(
        observed, synthetic, config_dict, station=inv,
        event=event, user_modules=user_modules,
        figure_mode=figure_mode, figure_dir=figure_dir,
        _verbose=_verbose)


class WindowASDF(ProcASDFBase):

    def __init__(self, path, param, verbose=False, debug=False):

        ProcASDFBase.__init__(self, path, param, verbose=verbose,
                              debug=debug)

    def _parse_param(self):

        # Get params
        param = self._parse_yaml(self.param)

        # reform the param from default
        default = param["default"]
        comp_settings = param["components"]
        results = {}

        if self.mpi_mode:

            # Get MPI rank
            myrank = self.comm.Get_rank()

            for _comp, _settings in comp_settings.items():
                if myrank == 0:
                    print("Preparing params for components: %s" % _comp)
                results[_comp] = deepcopy(default)
                if _settings is None:
                    continue
                for k, v in _settings.items():
                    if myrank == 0:
                        print("--> Modify key[%s] to value: %s --> %s"
                              % (k, results[_comp][k], v))
                    results[_comp][k] = v
        else:

            for _comp, _settings in comp_settings.items():

                results[_comp] = deepcopy(default)

                if _settings is None:
                    continue

                for k, v in _settings.items():
                    results[_comp][k] = v

        return results

    def _validate_path(self, path):
        necessary_keys = ["obsd_asdf", "obsd_tag", "synt_asdf", "synt_tag",
                          "output_file", "figure_mode", "figure_dir"]
        self._missing_keys(necessary_keys, path)

    def _validate_param(self, param):
        for key, value in param.items():
            necessary_keys = ["min_period", "max_period"]
            self._missing_keys(necessary_keys, value)
            minp = value["min_period"]
            maxp = value["max_period"]
            if minp > maxp:
                raise ValueError("min_period(%6.2f) is larger than max_period"
                                 "(%6.2f)" % (minp, maxp))

    def _core(self, path, param):

        obsd_file = path["obsd_asdf"]
        synt_file = path["synt_asdf"]
        output_file = path["output_file"]
        output_dir = os.path.dirname(output_file)

        self.check_input_file(obsd_file)
        self.check_input_file(synt_file)
        smart_mkdir(output_dir, mpi_mode=self.mpi_mode,
                    comm=self.comm)

        obsd_tag = path["obsd_tag"]
        synt_tag = path["synt_tag"]
        figure_mode = path["figure_mode"]
        figure_dir = path["figure_dir"]
        smart_mkdir(figure_dir, mpi_mode=self.mpi_mode,
                    comm=self.comm)

        obsd_ds = self.load_asdf(obsd_file)
        synt_ds = self.load_asdf(synt_file)

        event = obsd_ds.events[0]

        # Ridvan Orsvuran, 2016
        # take out the user module values
        user_modules = {}
        for key, value in param.items():
            user_modules[key] = value.pop("user_module", None)

        config_dict, instrument_merge_flag = load_window_config(param)

        winfunc = partial(window_wrapper, config_dict=config_dict,
                          obsd_tag=obsd_tag, synt_tag=synt_tag,
                          user_modules=user_modules,
                          event=event, figure_mode=figure_mode,
                          figure_dir=figure_dir, _verbose=self._verbose)

        # windows = \
        #     obsd_ds.process_two_files_no_mpi(synt_ds, winfunc)

        windows = \
            obsd_ds.process_two_files_without_parallel_output(synt_ds, winfunc)

        if instrument_merge_flag:
            # merge multiple instruments
            results = merge_windows(windows)
        else:
            # nothing is done
            results = windows

        # Take of the json and add stats to save the statistics to file
        stats_logfile = output_file[:-4] + "stats.json"

        # stats windows on rand 0
        stats_all_windows(results, obsd_tag, synt_tag,
                          instrument_merge_flag,
                          stats_logfile)

        # Write to file
        write_window_json(results, output_file)



"""
This script generates the adjoint stations from stations.json file and
measurements files. The output file is STATIONS_ADJOINT, which will
be used in the adjoint simulations later.
"""
from .utils import write_stations_file


def extract_usable_stations_from_one_period(measures):
    """ Extract usable stations and channels from measurements file """
    stations = []
    channels = []
    for sta, sta_info in measures.items():
        n_measure_sta = 0
        # append each usable channel
        for chan, chan_info in sta_info.items():
            n_measure_chan = len(chan_info)
            if n_measure_chan > 0:
                channels.append(chan)
            n_measure_sta += n_measure_chan
        # append each usable station
        if n_measure_sta > 0:
            stations.append(sta)

    return stations, channels


def extract_usable_stations_from_measurements(measurements):
    stations = set()
    channels = set()

    for period, measures in measurements.items():
        stations_one_period, channels_one_period = \
            extract_usable_stations_from_one_period(measures)
        print("[Period:%s]Number of stations and channels: %d, %d" %
              (period, len(stations_one_period), len(channels_one_period)))
        stations = stations.union(set(stations_one_period))
        channels = channels.union(set(channels_one_period))

    print("Total number of station and channels: %d, %d" %
          (len(stations), len(channels)))

    return stations, channels


def extract_one_station(chan, stations):
    nw, sta, loc, comp = chan.split(".")
    if comp[-1] == "Z":
        # directly get the station information
        info = stations[chan]
    else:
        # if horizontal components, first try BHE and then try BH1
        new_id_e = "%s.%s.%s.%sE" % (nw, sta, loc, comp[0:2])
        new_id_1 = "%s.%s.%s.%s1" % (nw, sta, loc, comp[0:2])
        if new_id_e in stations:
            info = stations[new_id_e]
        elif new_id_1 in stations:
            info = stations[new_id_1]
        else:
            raise ValueError("Can not locate station(%s) in staitons file"
                             % (chan))
    return info


def prepare_adjoint_station_information(usable_channels, stations):
    """ Based on usable channels, extract adjoint station information """
    adjoint_stations_info = {}

    for chan_id in usable_channels:
        nw, sta, loc, comp = chan_id.split(".")
        info = extract_one_station(chan_id, stations)
        sta_id = "%s.%s" % (nw, sta)
        if sta_id not in adjoint_stations_info:
            # if no previous, just add
            adjoint_stations_info[sta_id] = info
        else:
            # if previous, check if current is Z component
            if comp[-1] == "Z":
                adjoint_stations_info[sta_id] = info
            else:
                continue

    adjoint_stations = {}
    for sta_id, sta_info in adjoint_stations_info.items():
        adjoint_stations[sta_id] = [
            sta_info["latitude"], sta_info["longitude"],
            sta_info["elevation"], sta_info["depth"]]

    return adjoint_stations


def check_adjoint_stations_consistency(adjoint_stations, usable_stations):
    if len(adjoint_stations) != len(usable_stations):
        raise ValueError("Inconsistent between adjoint_stations and "
                         "usable_stations")

    set1 = set(adjoint_stations.keys())
    set2 = set(usable_stations)
    if set1 != set2:
        print("Stations more: %s" % (set1 - set2))
        print("Stations less: %s" % (set2 - set1))
        raise ValueError("Inconsistent between adjoint_stations and "
                         "usable_stations")

    print("Validation check passed")


def benchmark_stations(adjoint_stations):
    """
    Benchmark a few common stations(latitude and longitude)
    just to check if it is correct
    """
    threshold = 0.01

    def is_close(values, true_values):
        for _v1, _v2 in zip(values, true_values):
            if abs(_v1 - _v2) > threshold:
                return False
        return True

    true_values = {
        "II.AAK": [42.6375, 74.4942], "II.ABPO": [-19.0180, 47.2290],
        "II.EFI": [-51.6753, -58.0637], "IU.AFI": [-13.9093, -171.7773],
        "IU.ANMO": [34.9460, -106.4571], "G.CAN": [-35.3187, 148.9963]
    }

    npass = 0
    nfail = 0
    for key in true_values:
        if key not in adjoint_stations:
            continue
        if not is_close(adjoint_stations[key], true_values[key]):
            print("Fails at benchmark station %s" % key)
            nfail += 1
        else:
            npass += 1

    if nfail != 0:
        raise ValueError("Number of benchmark fails: %d" % nfail)

    return npass


def generate_adjoint_stations(measurements, stations, outputfn,
                              benchmark_flag=True):
    """
    This program takes in measurements and stations, and output
    the STATIONS_ADJOINT for stations has adjoint measurements

    :param measurements: dict contains measurements information
        from several period bands, generated by
        pypaw-measure_adjoint_asdf
    :type measurements: dict
    :param stations: dict contains station information, including
        station location and instrument information for all
        channels
    :type stations: dict
    :param outputfn: output STATIONS file
    :type outputfn: str
    :param benchmark_flag: whether benchmark some stations with
        standard locations values
    :type benchmark_flag: bool
    """
    usable_stations, usable_channels = \
        extract_usable_stations_from_measurements(measurements)

    adjoint_stations = prepare_adjoint_station_information(
        usable_channels, stations)

    check_adjoint_stations_consistency(adjoint_stations, usable_stations)

    if benchmark_flag:
        npass = benchmark_stations(adjoint_stations)
        print("Benchmark passed at level: %d" % npass)

    print("Write output station file in: %s" % outputfn)
    write_stations_file(adjoint_stations, outputfn)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

Source and Receiver classes of Instaseis.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2014
    Martin van Driel (Martin@vanDriel.de), 2014
    Wenjie Lei (lei@princeton.edu), 2016
    Lucas Sawade (lsawade@princeton.edu), 2019

:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lgpl.html)

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import numpy as np
from obspy import UTCDateTime, read_events
import warnings


class CMTSource(object):
    """
    Class to handle a seismic moment tensor source including a source time
    function.
    """
    def __init__(self, origin_time=UTCDateTime(0),
                 pde_latitude=0.0, pde_longitude=0.0, mb=0.0, ms=0.0,
                 pde_depth_in_m=None, region_tag=None, eventname=None,
                 cmt_time=0.0, half_duration=0.0, latitude=0.0, longitude=0.0,
                 depth_in_m=None, m_rr=0.0, m_tt=0.0, m_pp=0.0, m_rt=0.0,
                 m_rp=0.0, m_tp=0.0):
        """
        :param latitude: latitude of the source in degree
        :param longitude: longitude of the source in degree
        :param depth_in_m: source depth in m
        :param m_rr: moment tensor components in r, theta, phi in Nm
        :param m_tt: moment tensor components in r, theta, phi in Nm
        :param m_pp: moment tensor components in r, theta, phi in Nm
        :param m_rt: moment tensor components in r, theta, phi in Nm
        :param m_rp: moment tensor components in r, theta, phi in Nm
        :param m_tp: moment tensor components in r, theta, phi in Nm
        :param time_shift: correction of the origin time in seconds. only
            useful in the context of finite sources
        :param sliprate: normalized source time function (sliprate)
        :param dt: sampling of the source time function
        :param origin_time: The origin time of the source. This will be the
            time of the first sample in the final seismogram. Be careful to
            adjust it for any time shift or STF (de)convolution effects.
        """
        self.origin_time = origin_time
        self.pde_latitude = pde_latitude
        self.pde_longitude = pde_longitude
        self.pde_depth_in_m = pde_depth_in_m
        self.mb = mb
        self.ms = ms
        self.region_tag = region_tag
        self.eventname = eventname
        self.cmt_time = cmt_time
        self.half_duration = half_duration
        self.latitude = latitude
        self.longitude = longitude
        self.depth_in_m = depth_in_m
        self.m_rr = m_rr
        self.m_tt = m_tt
        self.m_pp = m_pp
        self.m_rt = m_rt
        self.m_rp = m_rp
        self.m_tp = m_tp

    @classmethod
    def from_CMTSOLUTION_file(cls, filename):
        """
        Initialize a source object from a CMTSOLUTION file.
        :param filename: path to the CMTSOLUTION file
        """

        with open(filename, "rt") as f:
            line = f.readline()
            origin_time = line[5:].strip().split()[:6]
            values = list(map(int, origin_time[:-1])) + \
                [float(origin_time[-1])]
            try:
                origin_time = UTCDateTime(*values)
            except (TypeError, ValueError):
                warnings.warn("Could not determine origin time from line: %s"
                              % line)
                origin_time = UTCDateTime(0)
            otherinfo = line[4:].strip().split()[6:]
            pde_lat = float(otherinfo[0])
            pde_lon = float(otherinfo[1])
            pde_depth_in_m = float(otherinfo[2]) * 1e3
            mb = float(otherinfo[3])
            ms = float(otherinfo[4])
            region_tag = ' '.join(otherinfo[5:])

            eventname = f.readline().strip().split()[-1]
            time_shift = float(f.readline().strip().split()[-1])
            cmt_time = origin_time + time_shift
            half_duration = float(f.readline().strip().split()[-1])
            latitude = float(f.readline().strip().split()[-1])
            longitude = float(f.readline().strip().split()[-1])
            depth_in_m = float(f.readline().strip().split()[-1]) * 1e3

            # unit: N/m
            m_rr = float(f.readline().strip().split()[-1])  # / 1e7
            m_tt = float(f.readline().strip().split()[-1])  # / 1e7
            m_pp = float(f.readline().strip().split()[-1])  # / 1e7
            m_rt = float(f.readline().strip().split()[-1])  # / 1e7
            m_rp = float(f.readline().strip().split()[-1])  # / 1e7
            m_tp = float(f.readline().strip().split()[-1])  # / 1e7

        return cls(origin_time=origin_time,
                   pde_latitude=pde_lat, pde_longitude=pde_lon, mb=mb, ms=ms,
                   pde_depth_in_m=pde_depth_in_m, region_tag=region_tag,
                   eventname=eventname, cmt_time=cmt_time,
                   half_duration=half_duration, latitude=latitude,
                   longitude=longitude, depth_in_m=depth_in_m,
                   m_rr=m_rr, m_tt=m_tt, m_pp=m_pp, m_rt=m_rt,
                   m_rp=m_rp, m_tp=m_tp)

    @classmethod
    def from_quakeml_file(cls, filename):
        """
        Initialiaze a source object from a quakeml file
        :param filename: path to a quakeml file
        """
        cat = read_events(filename)
        event = cat[0]
        for origin in event.origins:
            if origin.origin_type == 'centroid':
                cmtsolution = origin
            else:
                pdesolution = origin

        origin_time = pdesolution.time
        pde_lat = pdesolution.latitude
        pde_lon = pdesolution.longitude
        pde_depth_in_m = pdesolution.depth
        mb = 0.0
        ms = 0.0
        for mag in event.magnitudes:
            if mag.magnitude_type == "Mb":
                mb = mag.mag
            elif mag.magnitude_type == "MS":
                ms = mag.mag
        # Get region tag
        try:
            region_tag = cmtsolution.region
        except Exception:
            try:
                region_tag = cmtsolution.region
            except Exception:
                warnings.warn("Region tag not found.")

        for descrip in event.event_descriptions:
            if descrip.type == "earthquake name":
                eventname = descrip.text
            else:
                eventname = ""
        cmt_time = cmtsolution.time
        focal_mechanism = event.focal_mechanisms[0]
        half_duration = \
            focal_mechanism.moment_tensor.source_time_function.duration/2.0
        latitude = cmtsolution.latitude
        longitude = cmtsolution.longitude
        depth_in_m = cmtsolution.depth
        tensor = focal_mechanism.moment_tensor.tensor
        m_rr = tensor.m_rr * 1e7
        m_tt = tensor.m_tt * 1e7
        m_pp = tensor.m_pp * 1e7
        m_rt = tensor.m_rt * 1e7
        m_rp = tensor.m_rp * 1e7
        m_tp = tensor.m_tp * 1e7

        return cls(origin_time=origin_time,
                   pde_latitude=pde_lat, pde_longitude=pde_lon, mb=mb, ms=ms,
                   pde_depth_in_m=pde_depth_in_m, region_tag=region_tag,
                   eventname=eventname, cmt_time=cmt_time,
                   half_duration=half_duration, latitude=latitude,
                   longitude=longitude, depth_in_m=depth_in_m,
                   m_rr=m_rr, m_tt=m_tt, m_pp=m_pp, m_rt=m_rt,
                   m_rp=m_rp, m_tp=m_tp)

    def write_CMTSOLUTION_file(self, filename):
        """
        Initialize a source object from a CMTSOLUTION file.
        :param filename: path to the CMTSOLUTION file
        """
        with open(filename, "w") as f:
            # Reconstruct the first line as well as possible. All
            # hypocentral information is missing.
            f.write(
                " PDE %4i %2i %2i %2i %2i %5.2f %8.4f %9.4f %5.1f %.1f %.1f"
                " %s\n" % (
                    self.origin_time.year,
                    self.origin_time.month,
                    self.origin_time.day,
                    self.origin_time.hour,
                    self.origin_time.minute,
                    self.origin_time.second
                    + self.origin_time.microsecond / 1E6,
                    self.pde_latitude,
                    self.pde_longitude,
                    self.pde_depth_in_m / 1e3,
                    self.mb,
                    self.ms,
                    str(self.region_tag)))
            f.write('event name:     %s\n' % (str(self.eventname)))
            f.write('time shift:%12.4f\n' % (self.time_shift,))
            f.write('half duration:%9.4f\n' % (self.half_duration,))
            f.write('latitude:%14.4f\n' % (self.latitude,))
            f.write('longitude:%13.4f\n' % (self.longitude,))
            f.write('depth:%17.4f\n' % (self.depth_in_m / 1e3,))
            f.write('Mrr:%19.6e\n' % self.m_rr)  # * 1e7,))
            f.write('Mtt:%19.6e\n' % self.m_tt)  # * 1e7,))
            f.write('Mpp:%19.6e\n' % self.m_pp)  # * 1e7,))
            f.write('Mrt:%19.6e\n' % self.m_rt)  # * 1e7,))
            f.write('Mrp:%19.6e\n' % self.m_rp)  # * 1e7,))
            f.write('Mtp:%19.6e\n' % self.m_tp)  # * 1e7,))

    @property
    def M0(self):
        """
        Scalar Moment M0 in Nm
        """
        return (self.m_rr ** 2 + self.m_tt ** 2 + self.m_pp ** 2
                + 2 * self.m_rt ** 2 + 2 * self.m_rp ** 2
                + 2 * self.m_tp ** 2) ** 0.5 * 0.5 ** 0.5

    @property
    def moment_magnitude(self):
        """
        Moment magnitude M_w
        """
        return 2.0 / 3.0 * (np.log10(self.M0) - 7.0) - 6.0

    @property
    def time_shift(self):
        """
        Time shift between cmtsolution and pdesolution
        """
        return self.cmt_time - self.origin_time

    @property
    def tensor(self):
        """
        List of moment tensor components in r, theta, phi coordinates:
        [m_rr, m_tt, m_pp, m_rt, m_rp, m_tp]
        """
        return np.array([self.m_rr, self.m_tt, self.m_pp, self.m_rt, self.m_rp,
                         self.m_tp])

    def __str__(self):
        return_str = 'CMT Source -- %s\n' % self.eventname
        return_str += 'origin time(pde): %s\n' % self.origin_time
        return_str += 'pde location(lat, lon): %f, %f deg\n' \
            % (self.pde_latitude, self.pde_longitude)
        return_str += 'pde depth: %f\n' % self.pde_depth_in_m
        return_str += 'CMT time: %s\n' % self.cmt_time
        return_str += 'CMT location(lat, lon): %f, %f deg\n' \
            % (self.latitude, self.longitude)
        return_str += 'CMT depth: %6.1e km\n' \
                      % (self.depth_in_m / 1e3,)
        return_str += 'half duration: %f\n' % self.half_duration
        return_str += 'Moment Tensor [Mrr, Mtt, Mpp, Mrt, Mrp, Mtp]: \n'
        return_str += '              %s\n' \
            % self.tensor
        return_str += 'Magnitude: %4.2f(mw), %4.2f(mb), %4.2f(ms)\n' \
                      % (self.moment_magnitude, self.mb, self.ms)
        return_str += 'region tag: %s' % self.region_tag

        return return_str

    def __len__(self):
        return len(self.__dict__)

    def __iter__(self):
        """ Making the class iterable through key,value pairs. """
        # first start by grabbing the Class items
        iters = self.__dict__.items()

        # now 'yield' through the items
        for x, y in iters:
            yield x

    def __getitem__(self, item):
        """ Making the CMT Source subscriptable with indeces."""
        return self.__dict__[item]

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Parent class for singal processing asdf file and
handles parallel I/O so they are invisible to users.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
    Lucas Sawade (lsawade@princeton.edu) 2019
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
import sys
import json
import os
import time
import yaml
import matplotlib.pyplot as plt


def is_mpi_env():
    """
    Test if current environment is MPI or not
    """
    try:
        import mpi4py
    except ImportError:
        return False

    try:
        import mpi4py.MPI
    except ImportError:
        return False

    if mpi4py.MPI.COMM_WORLD.size == 1 and mpi4py.MPI.COMM_WORLD.rank == 0:
        return False

    return True


def _get_mpi_comm():
    from mpi4py import MPI
    return MPI.COMM_WORLD


def dump_json(content, filename):
    with open(filename, 'w') as fh:
        json.dump(content, fh, indent=2, sort_keys=True)


class JSONObject(object):
    def __init__(self, d):
        self.__dict__ = d


def read_json_file(parfile, obj_hook=False):
    """
    Hook json to an JSONObject instance
    """
    if not os.path.exists(parfile):
        raise ValueError("parfile not exists:%s" % parfile)
    with open(parfile, 'r') as f:
        if obj_hook:
            data = json.load(f, object_hook=JSONObject)
        else:
            data = json.load(f)
    return data


def smart_read_json(json_file, mpi_mode=True, comm=None, object_hook=False):
    """
    read json file under mpi and multi-processing environment.
    Hook it to an JSONObject(not the conventional way to
    read it as object)
    """
    if not mpi_mode:
        json_obj = read_json_file(json_file, obj_hook=object_hook)
    else:
        if comm is None:
            comm = _get_mpi_comm()
        rank = comm.Get_rank()
        if rank == 0:
            try:
                json_obj = read_json_file(json_file, obj_hook=object_hook)
            except Exception as err:
                print("Error in %s | error: %s" % (json_file, err))
                comm.Abort()
        else:
            json_obj = None
        json_obj = comm.bcast(json_obj, root=0)
    return json_obj


def write_yaml_file(d, filename, **kwargs):
    """Writes dictionary to given yaml file.

    Args:
          d: Dictionary to be written into the yaml file
          filename: string with filename of the file to be written.

    """
    with open(filename, 'w') as yaml_file:
        yaml.dump(d, yaml_file, default_flow_style=False, **kwargs)


def read_yaml_file(filename):
    with open(filename, "rb") as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


def smart_read_yaml(yaml_file, mpi_mode=True, comm=None):
    """
    Read yaml file into python dict, in mpi_mode or not
    """
    if not mpi_mode:
        yaml_dict = read_yaml_file(yaml_file)
    else:
        if comm is None:
            comm = _get_mpi_comm()
        rank = comm.rank
        if rank == 0:
            try:
                yaml_dict = read_yaml_file(yaml_file)
            except Exception as err:
                print("Error in read %s as yaml file: %s" % (yaml_file, err))
                comm.Abort()
        else:
            yaml_dict = None
        yaml_dict = comm.bcast(yaml_dict, root=0)
    return yaml_dict


def smart_check_path(path, mpi_mode=True, comm=None):
    if not mpi_mode:
        path_exists = os.path.exists(path)
    else:
        if comm is None:
            comm = _get_mpi_comm()
        rank = comm.rank
        if rank == 0:
            path_exists = os.path.exists(path)
        else:
            path_exists = None
        path_exists = comm.bcast(path_exists, root=0)
    return path_exists


def smart_remove_file(filename, mpi_mode=True, comm=None):
    if not smart_check_path(filename, mpi_mode=mpi_mode, comm=comm):
        return
    if mpi_mode:
        if comm is None:
            comm = _get_mpi_comm()
        rank = comm.rank
        comm.Barrier()
        if rank == 0:
            os.remove(filename)
        comm.Barrier()
    else:
        os.remove(filename)


def smart_mkdir(dirname, mpi_mode=True, comm=None):
    if os.path.exists(dirname):
        return
    if mpi_mode:
        if comm is None:
            comm = _get_mpi_comm()
        rank = comm.rank
        comm.Barrier()
        if rank == 0:
            os.makedirs(dirname)
        comm.Barrier()
    else:
        os.makedirs(dirname)


def drawProgressBar(percent, user_text="", barLen=20):
    """
    Draw status progress bars in terminal.
    Not recommendded used for job scripts(will create
    thousands of lines of output)
    """

    if user_text == "":
        user_text = "Status"

    sys.stdout.write("\r")
    progress = ""
    for i in range(barLen):
        if i < int(barLen * percent):
            progress += "="
        else:
            progress += " "
    sys.stdout.write(
        "%s: [ %s ] %.2f%%" % (user_text, progress, percent * 100))
    sys.stdout.flush()

    if percent >= 1.0:
        sys.stdout.write("\n")


def timing(f):
    def wrap(*args, **kwargs):
        time1 = time.time()
        ret = f(*args, **kwargs)
        time2 = time.time()
        print('%s function took %0.3f s' % (f.__name__, (time2-time1)))
        return ret
    return wrap


def isclose(a, b, rel_tol=1.0e-09, abs_tol=0.0):
    """
    Compare two numbers see if they are close to each other to a
    tolerance level
    """
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def load_json(filename):
    with open(filename, 'r') as fh:
        return json.load(fh)


def load_yaml(filename):
    with open(filename, 'r') as fh:
        return yaml.load(fh)


def dump_yaml(content, filename):
    with open(filename, 'w') as fh:
        yaml.dump(content, fh, indent=2)


def reset_matplotlib():
    plt.switch_backend('agg')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

This script finallay inverts the CMT Solution.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""

from gcmt3d.workflow.tensor_rise import gradient

import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='cmt_file',
                        required=True, help="Path to CMT file in database")
    parser.add_argument('-p', action='store', dest='param_path', required=True,
                        help="Path to Parameter Directory")
    args = parser.parse_args()

    gradient(args.cmt_file, args.param_path)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script extract channel instrument type from obspy.inventory

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""

from __future__ import print_function, division, absolute_import
import os
from collections import defaultdict
from obspy import Inventory
from obspy import read_inventory


def safe_load_staxml(staxmlfile):
    try:
        inv = read_inventory(staxmlfile)
    except Exception as exp:
        raise("Failed to parse staxml file(%s) due to: %s"
              % (staxmlfile, exp))
    return inv


def extract_staxml_info(staxml):
    """ extract information from staionxml file or obspy.Inventory """
    instruments = defaultdict(dict)

    if isinstance(staxml, Inventory):
        inv = staxml
    else:
        if os.path.isfile(staxml):
            inv = safe_load_staxml(staxml)
        else:
            raise ValueError("Input staxml is neither obspy.Inventory or "
                             "staxml file")
    for nw in inv:
        nw_code = nw.code
        for sta in nw:
            sta_code = sta.code
            for chan in sta:
                chan_code = chan.code
                loc_code = chan.location_code
                key = "%s.%s.%s.%s" % (nw_code, sta_code, loc_code, chan_code)
                instruments[key]["latitude"] = chan.latitude
                instruments[key]["longitude"] = chan.longitude
                instruments[key]["elevation"] = chan.elevation
                instruments[key]["depth"] = chan.depth
                if chan.sensor.description is not None:
                    sensor_type = chan.sensor.description
                elif chan.sensor.type is not None:
                    sensor_type = chan.sensor.type
                else:
                    sensor_type = "None"
                instruments[key]["sensor"] = sensor_type

    return instruments



"""

This script contains binary to create inversion dictionaries.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: June 2019

"""

from gcmt3d.workflow.shape_inversion_dictionaries import inversion_dictionaries
import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='cmt_file',
                        required=True, help="Path to CMT file in database")
    parser.add_argument('-p', action='store', dest='param_path', required=True,
                        help="Path to Parameter Directory")
    args = parser.parse_args()

    inversion_dictionaries(args.cmt_file, args.param_path)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Filter the window based on the sensor type. For example, in long
period band(90-250s), we want to keep only STS-1 instrument windows.
For the input file, it requires 1) sensor type as json file; 2) windows
as json file. For the output, it is going to replace the origin window file
and keep a copy of original windows as "``***``.origin.json"

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016

:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""
from __future__ import (absolute_import, division, print_function)
from pprint import pprint
from copy import deepcopy
import numpy as np


def is_right_sensor(sensor, sensor_types):
    """
    Check if sensor in the list of sensor_types
    """
    for stype in sensor_types:
        if stype in sensor:
            return True
    return False


def count_windows(windows):
    """
    Count the number of channels and windows
    :param windows:
    :return:
    """
    nchans = 0
    nwins = 0
    nwins_comp = {}
    for stainfo in windows.values():
        for chan, chaninfo in stainfo.items():
            comp = chan.split(".")[-1][-1]
            _nw = len(chaninfo)
            if _nw == 0:
                continue
            nchans += 1
            nwins += _nw
            if comp not in nwins_comp:
                nwins_comp[comp] = 0
            nwins_comp[comp] += _nw

    return nchans, nwins, nwins_comp


def print_window_filter_summary(old_windows, new_windows):
    nchans_old, nwins_old, nwins_comp_old = count_windows(old_windows)
    nchans_new, nwins_new, nwins_comp_new = count_windows(new_windows)
    if nchans_old:
        nchans_rej = (nchans_old - nchans_new) / nchans_old
    else:
        nchans_rej = 1.0
    if nwins_old:
        nwins_rej = (nwins_old - nwins_new) / nwins_old
    else:
        nwins_rej = 1.0
    print("-" * 10)
    print("Number of channels old and new: %d --> %d(rej: %.2f %%)"
          % (nchans_old, nchans_new, nchans_rej * 100))
    print("Number of windows old and new:  %d --> %d(rej: %.2f %%)"
          % (nwins_old, nwins_new, nwins_rej * 100))
    print("Old Component window counts: %s" % (nwins_comp_old))
    print("New Component window counts: %s" % (nwins_comp_new))
    return {"nchannels_old": nchans_old, "nwindows_old": nwins_old,
            "nchannels_new": nchans_new, "nwindows_new": nwins_new,
            "channel_rejection_percentage": '%.2f' % (nchans_rej * 100),
            "window_rejection_percentage": '%.2f' % (nwins_rej * 100)}


def filter_windows_on_sensors(windows, stations, sensor_types, verbose=False):
    """
    Filter the windows based on sensor types and station information.
    Only sensor types in 'sensor_types' will be kept.
    """
    new_wins = {}
    print("sensor_types: %s" % sensor_types)

    if verbose:
        print("channel name" + " " * 13 + "|" + " " * 30
              + "sensor type | pick flag | wins | sum(chans, wins) |")

    total_chans = 0
    total_wins = 0
    for sta, sta_info in windows.items():
        sta_wins = {}
        for chan, chan_info in sta_info.items():
            pick_flag = False
            if len(chan_info) == 0:
                # if number of windows is 0
                continue
            try:
                # since windows are on RTZ component and
                # instruments are on NEZ compoennt, so
                # just use Z component instrument information
                zchan = chan[:-1] + "Z"
                _st = stations[zchan]["sensor"]
            except Exception:
                continue
            if is_right_sensor(_st, sensor_types):
                sta_wins[chan] = chan_info
                pick_flag = True
                total_chans += 1
                total_wins += len(chan_info)
            if verbose:
                print("channel(%15s) | %40s |%11s|%9d| (%d, %d)"
                      % (chan, _st[:40], pick_flag, len(chan_info),
                         total_chans, total_wins))

        if len(sta_wins) > 0:
            new_wins[sta] = sta_wins

    return new_wins


def get_measurements_std(measurements):
    """
    Calculate the mean and standard deviation values from the measurements
    """
    comp_dt_meas = {}
    comp_dlna_meas = {}

    for sta_info in measurements.values():
        for chan, chan_info in sta_info.items():
            comp = chan.split(".")[-1][-1]

            dts = [v["dt"] for v in chan_info]
            dlnas = [v["dlna"] for v in chan_info]

            if comp not in comp_dt_meas:
                comp_dt_meas[comp] = []
            comp_dt_meas[comp].extend(dts)

            if comp not in comp_dlna_meas:
                comp_dlna_meas[comp] = []
            comp_dlna_meas[comp].extend(dlnas)

    dt_means = dict((k, np.mean(v)) for (k, v) in comp_dt_meas.items())
    dt_stds = dict((k, np.std(v)) for (k, v) in comp_dt_meas.items())

    dlna_means = dict((k, np.mean(v)) for (k, v) in comp_dlna_meas.items())
    dlna_stds = dict((k, np.std(v)) for (k, v) in comp_dlna_meas.items())

    print("means of dt measurements:")
    pprint(dt_means)
    print("std of dt measurements:")
    pprint(dt_stds)
    print("means of dlna measurements:")
    pprint(dlna_means)
    print("std of dlna measurements:")
    pprint(dlna_stds)

    return dt_means, dt_stds, dlna_means, dlna_stds


def get_user_bound(info):
    """ user specified bound for measurements filter """
    v = [info["tshift_reference"] - info["tshift_acceptance_level"],
         info["tshift_reference"] + info["tshift_acceptance_level"],
         info["dlna_reference"] - info["dlna_acceptance_level"],
         info["dlna_reference"] + info["dlna_acceptance_level"]]
    if v[0] > v[1] or v[2] > v[3]:
        raise ValueError("error on user bound: %s" % v)
    return v


def get_std_bound(mean, std, std_ratio):
    """ bound from mean and std values from measurements """
    b = [mean - std_ratio * std, mean + std_ratio * std]
    if b[0] > b[1]:
        raise ValueError("Error on std bound: %s" % b)
    return b


def filter_measurements_on_bounds(windows, measurements, bounds,
                                  comp_keep_flag=None):
    """
    filter the windows based on its measurements
    """
    def _filter_(chan_wins, chan_meas, _bounds):
        """ filter one channel window and measurements """
        new_wins = []
        new_meas = []
        if len(chan_wins) != len(chan_meas):
            raise ValueError("number of windows is not the same as number"
                             "measurements")
        for win, meas in zip(chan_wins, chan_meas):
            if (meas["dt"] >= _bounds["dt"][0]) \
                    and (meas["dt"] <= _bounds["dt"][1]) \
                    and (meas["dlna"] >= _bounds["dlna"][0]) \
                    and (meas["dlna"] <= _bounds["dlna"][1]):
                new_wins.append(deepcopy(win))
                new_meas.append(deepcopy(meas))

        if len(new_wins) != len(new_meas):
            raise ValueError("length of new_wins(%d) and new_meas(%d) not "
                             "the same!" % (len(new_wins), len(new_meas)))
        return new_wins, new_meas

    new_wins = {}
    new_meas = {}
    for sta, sta_info in windows.items():
        new_sta_wins = {}
        new_sta_meas = {}
        for chan, chan_info in sta_info.items():
            if len(chan_info) == 0:
                continue
            comp = chan.split(".")[-1][-1]
            if comp_keep_flag is not None and not comp_keep_flag[comp]:
                # throw comp if not keep
                continue
            m = measurements[sta][chan]
            w_f, m_f = _filter_(chan_info, m, bounds[comp])
            if len(w_f) == 0:
                continue
            new_sta_wins[chan] = w_f
            new_sta_meas[chan] = m_f

        if len(new_sta_wins) > 0:
            new_wins[sta] = new_sta_wins
            new_meas[sta] = new_sta_meas

    return new_wins, new_meas


def get_component_keep_flag(dt_means, dt_stds, dlna_means, dlna_stds,
                            comp_config):
    """
    Check if the mean and std of dt and dlna are within the range
    provided by the user. The default values are true. If the user
    doesn't provide any parameters, then all values will be kept
    as true
    """
    def _check_all_parameters_valid(_config):
        ts_range = _config["tshift_mean_range"]
        if ts_range[0] >= ts_range[1]:
            raise ValueError("tshift_mean_range error: %s" % ts_range)
        ts_std_level = _config["tshift_std_level"]
        if ts_std_level <= 0:
            raise ValueError("tshift_std_level(%f) <= 0" % ts_std_level)
        dlna_range = _config["dlna_mean_range"]
        if dlna_range[0] >= dlna_range[1]:
            raise ValueError("dlna_mean_range error: %s" % dlna_range)
        dlna_std_level = _config["dlna_std_level"]
        if dlna_std_level <= 0:
            raise ValueError("dlna_std_level(%f) <= 0" % dlna_std_level)

    def _check_in_range(v, vrange):
        if v > vrange[0] and v < vrange[1]:
            return True
        else:
            return False

    flags = {}
    for comp in dt_means:
        flags[comp] = True

    print("-" * 20)
    for comp in dt_means:
        if "tshift_mean_range" not in comp_config[comp]:
            # skip if user does not provide the parameter
            print("Not filtering on mean and std on component(%s) since"
                  "no user parameter(tshift_mean_range) provided" % comp)
            continue
        _config = comp_config[comp]
        _check_all_parameters_valid(_config)
        if not _check_in_range(dt_means[comp], _config["tshift_mean_range"]):
            print("Reject component(%s) due to dt mean(%f) out of range: %s"
                  % (comp, dt_means[comp], _config["tshift_mean_range"]))
            flags[comp] = False
            continue

        if dt_stds[comp] > _config["tshift_std_level"]:
            print("Reject component(%s) due to dt std(%f) larger than "
                  "threshold %f"
                  % (comp, dt_stds[comp], _config["tshift_std_level"]))
            flags[comp] = False
            continue

        if not _check_in_range(dlna_means[comp], _config["dlna_mean_range"]):
            print("Reject component(%s) due to dlna mean(%f) out of range: %s"
                  % (comp, dlna_means[comp], _config["dlna_mean_range"]))
            flags[comp] = False
            continue

        if dlna_stds[comp] > _config["dlna_std_level"]:
            print("Reject component(%s) due to dlna std(%f) larger than "
                  "threshold %f"
                  % (comp, dlna_stds[comp], _config["dlna_std_level"]))
            flags[comp] = False
            continue

    pprint("component keep flag:")
    pprint(flags)
    return flags


def get_measurement_final_bounds(
        comp_config, dt_means, dt_stds, dlna_means, dlna_stds):
    # get the measurement bounds
    final_bounds = {}
    for comp in dt_means:
        print("-" * 20 + "\nComponent: %s" % comp)
        user_bound = get_user_bound(comp_config[comp])
        dt_std_bound = get_std_bound(dt_means[comp], dt_stds[comp],
                                     comp_config[comp]["std_ratio"])
        dlna_std_bound = get_std_bound(dlna_means[comp], dlna_stds[comp],
                                       comp_config[comp]["std_ratio"])
        dt_bound = [max(dt_std_bound[0], user_bound[0]),
                    min(dt_std_bound[1], user_bound[1])]
        dlna_bound = [max(dlna_std_bound[0], user_bound[2]),
                      min(dlna_std_bound[1], user_bound[3])]
        final_bounds[comp] = {"dt": dt_bound, "dlna": dlna_bound}
        print("user specified bound [dt_min, dt_max, dlna_min, dlna_max]: %s"
              % user_bound)
        print("std bound dt: %s -- dlna: %s" % (dt_std_bound, dlna_std_bound))
        print("final dt bound: %s" % dt_bound)
        print("final dlna bound: %s" % dlna_bound)

    return final_bounds


def filter_windows_on_measurements(windows, measurements, measure_config):
    """
    Filter windows based on measurements and threshold specified by
    the user.
    """
    # calculate standard deviation for each component
    pprint("Config:")
    pprint(measure_config)

    if len(windows) == 0:
        return {}, {}

    dt_means, dt_stds, dlna_means, dlna_stds = \
        get_measurements_std(measurements)

    comp_config = measure_config['component']
    comp_keep_flag = get_component_keep_flag(
        dt_means, dt_stds, dlna_means, dlna_stds, comp_config)

    final_bounds = get_measurement_final_bounds(
        comp_config, dt_means, dt_stds, dlna_means, dlna_stds)

    new_wins, new_meas = filter_measurements_on_bounds(
        windows, measurements, final_bounds, comp_keep_flag=comp_keep_flag)

    return new_wins, new_meas


def check_consistency(windows, measurements):
    for sta, stainfo in windows.items():
        for chan, chan_info in stainfo.items():
            nwins_chan = len(chan_info)
            if nwins_chan == 0:
                continue
            try:
                nwins_meas = len(measurements[sta][chan])
            except KeyError as errmsg:
                raise KeyError("Missing %s(%s) in measurements: %s"
                               % (chan, sta, errmsg))
            if nwins_chan != nwins_meas:
                raise ValueError(
                    "Inconsistent between windows(%d) and measurements(%d)"
                    "from %s" % (nwins_chan, nwins_meas, chan))
    print("Pass the consistency check...")


def filter_windows(windows, stations, measurements, config, verbose=False):
    """
    Filter windows based on measurements and station information
    """
    check_consistency(windows, measurements)

    log = {}
    sensor_config = config["sensor"]
    if sensor_config["flag"]:
        print("=" * 10 + "  Filter on sensors  " + "=" * 10)
        windows_sensor = filter_windows_on_sensors(
            windows, stations, sensor_config["sensor_types"],
            verbose=verbose)
    else:
        windows_sensor = windows
    log["sensor"] = print_window_filter_summary(windows, windows_sensor)

    measure_config = config["measurement"]
    if measure_config["flag"]:
        print("=" * 10 + "  Filter on measurements  " + "=" * 10)
        windows_measure, measure_filtered = filter_windows_on_measurements(
            windows_sensor, measurements, measure_config)
    else:
        windows_measure = windows_sensor
        measure_filtered = measurements
    log["measurement"] = print_window_filter_summary(windows_sensor,
                                                     windows_measure)

    return windows_measure, measure_filtered, log



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This is a script that will create a database entry given a cmt solution in the


:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""

import os
from gcmt3d.data.management.skeleton import DataBaseSkeleton
from gcmt3d.asdf.utils import smart_read_yaml, is_mpi_env


def create_entry(cmt_filename, param_path):
    # Define parameter directory
    databaseparam_path = os.path.join(param_path,
                                      "Database/DatabaseParameters.yml")
    specfemspec_path = os.path.join(param_path,
                                    "SpecfemParams/SpecfemParams.yml")
    stations_path = os.path.join(param_path,
                                 "RequestParams/STATIONS")

    # Load Parameters
    DB_params = smart_read_yaml(databaseparam_path, mpi_mode=is_mpi_env())
    specfemspecs = smart_read_yaml(specfemspec_path, mpi_mode=is_mpi_env())

    # Check whether stationsfile in Parampath
    if os.path.exists(stations_path):
        stations_file = stations_path
    else:
        # if no stations file in the parameter directory,
        # the standard stations file is going to be used
        stations_file = None

    # Database Setup.
    DB = DataBaseSkeleton(basedir=DB_params["databasedir"],
                          cmt_fn=cmt_filename,
                          specfem_dir=specfemspecs["SPECFEM_DIR"],
                          stations_file=stations_file,
                          overwrite=DB_params['overwrite'])

    # Database Create entry
    DB.create_all()

    # Return new earthquake location.
    cmt_in_database = os.path.join(DB.Cdirs[0], "C" + DB.Cids[0])

    return cmt_in_database



"""

This script contains functions to create path files for the processing of the
observed and synthetic data.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: June 2019

"""

import os
import glob
import yaml

import logging
from ...log_util import modify_logger

# Create logger
logger = logging.getLogger(__name__)
modify_logger(logger)


def is_mpi_env():
    """
    Test if current environment is MPI or not
    """
    try:
        import mpi4py
    except ImportError:
        return False

    try:
        import mpi4py.MPI
    except ImportError:
        return False

    if mpi4py.MPI.COMM_WORLD.size == 1 and mpi4py.MPI.COMM_WORLD.rank == 0:
        return False

    return True


def _get_mpi_comm():
    from mpi4py import MPI
    return MPI.COMM_WORLD


def write_yaml_file(d, filename, **kwargs):
    """Writes dictionary to given yaml file.

    Args:
          d: Dictionary to be written into the yaml file
          filename: string with filename of the file to be written.

    """
    with open(filename, 'w') as yaml_file:
        yaml.dump(d, yaml_file, default_flow_style=False, **kwargs)


def read_yaml_file(filename):
    with open(filename) as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


def smart_read_yaml(yaml_file, mpi_mode=True, comm=None):
    """
    Read yaml file into python dict, in mpi_mode or not
    """
    if not mpi_mode:
        yaml_dict = read_yaml_file(yaml_file)
    else:
        if comm is None:
            comm = _get_mpi_comm()
        rank = comm.rank
        if rank == 0:
            try:
                yaml_dict = read_yaml_file(yaml_file)
            except Exception as err:
                print("Error in read %s as yaml file: %s" % (yaml_file, err))
                comm.Abort()
        else:
            yaml_dict = None
        yaml_dict = comm.bcast(yaml_dict, root=0)
    return yaml_dict


attr = ["CMT", "CMT_rr", "CMT_tt", "CMT_pp", "CMT_rt", "CMT_rp", "CMT_tp",
        "CMT_depth", "CMT_lat", "CMT_lon"]


def create_process_path_obs(cmt_filename, process_dir):
    """ This function writes a yaml conversion path file for 1 simulation
    file. This file is later on need for the creation of ASDF files and the
    processing involved ASDF files.

    :param cmt_filename: cmt_filename in the database (Important this only
                         works if the directory structure exists already)
    :param process_dir: path to the directory containing all processing files.
    :param verbose: boolean on whether process info should be written

    """

    # CMT directory Name
    cmt_dir = os.path.dirname(os.path.abspath(cmt_filename))

    # Output directory
    process_path_dir = os.path.join(cmt_dir, "seismograms",
                                    "process_paths")

    # Get all process possibilities
    process_param_files = glob.glob(os.path.join(process_dir, "*"))

    logger.verbose("Processing parameter files to be used:")
    logger.verbose(" ")
    for file in process_param_files:
        logger.verbose("    " + file)
    logger.verbose(" ")

    for _i, process_param_file in enumerate(process_param_files):

        # Get band
        process_params = smart_read_yaml(process_param_file,
                                         mpi_mode=is_mpi_env())
        band = process_params["pre_filt"][1:-1]
        lP = 1 / band[1]  # Get low period bound from hz filtervalue
        hP = 1 / band[0]  # Get high period bound from hz filtervalue

        # Define input ASDF
        input_asdf = os.path.join(cmt_dir, "seismograms", "obs",
                                  "raw_observed.h5")
        input_tag = "obs"

        # Output file parameters
        output_asdf = os.path.join(cmt_dir, "seismograms",
                                   "processed_seismograms",
                                   "processed_observed.%03.0f_%03.0f.h5"
                                   % (lP, hP))
        output_tag = "%03.0f_%03.0f_obsd" % (lP, hP)

        # Pathfile directory
        yaml_file_path = os.path.join(process_path_dir,
                                      "process_observed.%03.0f_%03.0f.yml"
                                      % (lP, hP))

        # Create dictionary
        logger.verbose("Writing path file %s." % yaml_file_path)

        d = {"input_asdf": input_asdf,
             "input_tag": input_tag,
             "output_asdf": output_asdf,
             "output_tag": output_tag,
             "process_param_file": process_param_file}

        # Writing the directory to file
        write_yaml_file(d, yaml_file_path)


def create_process_path_syn(cmt_filename, process_dir, npar, verbose=True):
    """ This function writes a yaml processing path file all simulations
    file. This is needed for the processing of the ASDF files.


    :param cmt_filename: cmt_filename in the database (Important this only
        works if the directory structure exists already)
    :param process_dir: path to the directory containing all processing files.
    :param npar: number of parameters to invert for
    :param verbose: boolean on whether process info should be written
    """

    # CMT directory Name
    cmt_dir = os.path.dirname(os.path.abspath(cmt_filename))

    # Output directory
    process_path_dir = os.path.join(cmt_dir, "seismograms",
                                    "process_paths")

    # Get all process possibilities
    process_param_files = glob.glob(os.path.join(process_dir, "*"))

    logger.verbose("Processing parameter files to be used:")
    logger.verbose(" ")
    for file in process_param_files:
        logger.verbose("    " + file)
    logger.verbose(" ")

    for _i, process_param_file in enumerate(process_param_files):

        for _j, at in enumerate(attr[:npar + 1]):

            # Get band
            process_params = smart_read_yaml(process_param_file,
                                             mpi_mode=is_mpi_env())
            band = process_params["pre_filt"][1:-1]
            lP = 1 / band[1]  # Get low period bound from hz filtervalue
            hP = 1 / band[0]  # Get high period bound from hz filtervalue

            # Define input ASDF
            input_asdf = os.path.join(cmt_dir, "seismograms", "syn",
                                      "%s.h5" % at)
            input_tag = "syn"

            # Output file parameters
            output_asdf = os.path.join(cmt_dir, "seismograms",
                                       "processed_seismograms",
                                       'processed_synthetic_'
                                       '%s.%03.0f_%03.0f.h5'
                                       % (at, lP, hP))
            output_tag = "%03.0f_%03.0f_synt" % (lP, hP)

            # Pathfile directory
            yaml_file_path = os.path.join(process_path_dir,
                                          "process_synthetic_"
                                          "%s.%03.0f_%03.0f.yml"
                                          % (at, lP, hP))

            # Create dictionary
            logger.verbose("Writing path file %s." % yaml_file_path)

            d = {"input_asdf": input_asdf,
                 "input_tag": input_tag,
                 "output_asdf": output_asdf,
                 "output_tag": output_tag,
                 "process_param_file": process_param_file}

            # Writing the directory to file
            write_yaml_file(d, yaml_file_path)


def create_window_path(cmt_filename, window_process_dir,
                       figure_mode=False, verbose=True):
    """ This function writes a yaml processing path file all simulations
    file. This is needed for the processing of the ASDF files.

    :param cmt_filename: cmt_filename in the database (Important this only
                        works if the directory structure exists already)
    :param process_dir: path to the directory containing all processing files.
    :param npar: number of parameters to invert for
    :param verbose: boolean on whether process info should be written
    """

    # CMT directory Name
    cmt_dir = os.path.dirname(os.path.abspath(cmt_filename))

    # Output directory
    window_path_dir = os.path.join(cmt_dir, "window_data", "window_paths")

    # Get all process possibilities
    window_param_files = glob.glob(os.path.join(window_process_dir, "*"))

    logger.info("Window parameter files to be used:")
    logger.info("--------------------------------")
    logger.info(" ")
    for file in window_param_files:
        logger.info("   " + file)
    logger.info(" ")

    for _i, window_param_file in enumerate(window_param_files):

        # Get band
        window_process_params = smart_read_yaml(window_param_file,
                                                mpi_mode=is_mpi_env())
        # Get lower and upper period bound
        lP = window_process_params["default"]["min_period"]
        hP = window_process_params["default"]["max_period"]

        # Check BodyWave/SurfaceWave flag
        if "body" in os.path.basename(window_param_file):
            wave_type = "#body_wave"
        elif "surface" in os.path.basename(window_param_file):
            wave_type = "#surface_wave"
        else:
            wave_type = ""

        # Define observed ASDF
        obsd_asdf = os.path.join(cmt_dir, "seismograms",
                                 "processed_seismograms",
                                 "processed_observed.%03.0f_%03.0f.h5"
                                 % (lP, hP))
        obsd_tag = "%03.0f_%03.0f_obsd" % (lP, hP)

        # Synthetic ASDF
        synt_asdf = os.path.join(cmt_dir, "seismograms",
                                 "processed_seismograms",
                                 "processed_synthetic_CMT.%03.0f_%03.0f.h5"
                                 % (lP, hP))
        synt_tag = "%03.0f_%03.0f_synt" % (lP, hP)

        # Output file parameters
        output_file = os.path.join(cmt_dir, "window_data",
                                   'windows.%03.0f_%03.0f%s.json'
                                   % (lP, hP, wave_type))
        figure_dir = os.path.join(cmt_dir, "window_data",
                                  'window_plots.%03.0f_%03.0f%s'
                                  % (lP, hP, wave_type))

        # Pathfile directory
        yaml_file_path = os.path.join(window_path_dir,
                                      "windows.%03.0f_%03.0f%s.yml"
                                      % (lP, hP, wave_type))

        # Create dictionary
        logger.verbose("Writing path file %s." % yaml_file_path)

        d = {"obsd_asdf": obsd_asdf,
             "obsd_tag": obsd_tag,
             "synt_asdf": synt_asdf,
             "synt_tag": synt_tag,
             "output_file": output_file,
             "figure_mode": figure_mode,
             "figure_dir": figure_dir,
             "window_param_file": window_param_file}

        # Writing the directory to file
        write_yaml_file(d, yaml_file_path)


def get_processing_list(cmt_file_db, process_obs_dir, process_syn_dir, npar=9):
    """This function returns a list of all process path files. It is needed
    for the EnTK workflow. This way the processing of each ASDF file can be
    assigned to one task.

    :param cmt_file_db: the cmtsolution file in the database.
    :type cmt_file_db: str
    :param process_obs_dir: directory with the process parameter files for
                            the observed data
    :type cmt_file_db: str
    :param process_syn_dir: directory with the process parameter files for
                            the synthetic data
    :type process_syn_dir: str
    :param npar: Number of parameters to invert for
    :type npar: int
    :param verbose: verbose output if true
    :type verbose: bool
    :return: tuple of 3 lists - 1 all processing path files;
                                2 the observed output files;
                                3 the synthetic output files
    """

    # CMT directory Name
    cmt_dir = os.path.dirname(os.path.abspath(cmt_file_db))

    # Output directory
    process_path_dir = os.path.join(cmt_dir, "seismograms",
                                    "process_paths")

    # Get all process possibilities
    process_obs_param_files = glob.glob(os.path.join(process_obs_dir, "*"))
    process_syn_param_files = glob.glob(os.path.join(process_syn_dir, "*"))

    # If verbose print all used observed and synthetic files
    logger.info("Processing parameter files to be used:")
    logger.info("--------------------------------------")
    logger.info(" ")
    logger.info("  Observed:")
    for process_file in process_obs_param_files:
        logger.info("    " + process_file)
    logger.info("  Synthetic:")
    for process_file in process_syn_param_files:
        logger.info("    " + process_file)
    logger.info(" ")

    # Create empty process_path file list
    process_path_file_list = []
    obs_output_file_list = []
    syn_output_file_list = []

    # Get observed path files
    for _i, process_param_file in enumerate(process_obs_param_files):
        # Get band
        process_params = smart_read_yaml(process_param_file,
                                         mpi_mode=is_mpi_env())
        band = process_params["pre_filt"][1:-1]
        lP = 1 / band[1]  # Get low period bound from hz filtervalue
        hP = 1 / band[0]  # Get high period bound from hz filtervalue

        # Output file parameters
        obs_output_file_list.append(os.path.join(cmt_dir, "seismograms",
                                                 "processed_seismograms",
                                                 "processed_observed"
                                                 ".%03.0f_%03.0f.h5"
                                                 % (lP, hP)))

        # Pathfile directory
        yaml_file_path = os.path.join(process_path_dir,
                                      "process_observed.%03.0f_%03.0f.yml"
                                      % (lP, hP))
        # Add the path file to list
        process_path_file_list.append(yaml_file_path)

    # Get synthetic path files
    for _i, process_param_file in enumerate(process_syn_param_files):
        for _j, at in enumerate(attr[:npar + 1]):
            # Get band
            process_params = smart_read_yaml(process_param_file,
                                             mpi_mode=is_mpi_env())
            band = process_params["pre_filt"][1:-1]
            lP = 1 / band[1]  # Get low period bound from hz filtervalue
            hP = 1 / band[0]  # Get high period bound from hz filtervalue

            # Output file parameters
            syn_output_file_list.append(os.path.join(cmt_dir, "seismograms",
                                                     "processed_seismograms",
                                                     'processed_synthetic_'
                                                     '%s.%03.0f_%03.0f.h5'
                                                     % (at, lP, hP)))

            # Pathfile directory
            yaml_file_path = os.path.join(process_path_dir,
                                          "process_synthetic_"
                                          "%s.%03.0f_%03.0f.yml"
                                          % (at, lP, hP))

            # Add the path file to list
            process_path_file_list.append(yaml_file_path)

    logger.info("Processing Path files:")
    logger.info("--------------------------------")
    logger.info(" ")
    for process_file in process_path_file_list:
        logger.info("    " + process_file)
    logger.info(" ")

    return process_path_file_list, obs_output_file_list, syn_output_file_list


def get_windowing_list(cmt_file_db, window_process_dir):
    """This function returns a list of all windowing path files. It is needed
    for the EnTK workflow. This way the windowing of each passband ASDF file
    can be assigned to one task.

    :param cmt_file_db: the cmtsolution file in the database.
    :type cmt_file_db: str
    :param window_process_dir: path to the process directory
    :type window_process_dir: str
    :param verbose: if True verbose output
    :type verbose: bool

    :return:  tuple of 2 lists - 1 all windowing path files; 2 the output files
    """

    # CMT directory Name
    cmt_dir = os.path.dirname(os.path.abspath(cmt_file_db))

    # Output directory
    window_path_dir = os.path.join(cmt_dir, "window_data", "window_paths")

    # Get all process possibilities
    window_param_files = glob.glob(os.path.join(window_process_dir, "*"))

    logger.info("Windowing parameter files to be used:")
    logger.info("--------------------------------------")
    logger.info(" ")
    for param_file in window_param_files:
        logger.info(param_file)
    logger.info(" ")

    window_processing_list = []
    output_file_list = []

    for _i, window_param_file in enumerate(window_param_files):

        # Get band
        window_process_params = smart_read_yaml(window_param_file,
                                                mpi_mode=is_mpi_env())
        # Get lower and upper period bound
        lP = window_process_params["default"]["min_period"]
        hP = window_process_params["default"]["max_period"]

        # Check BodyWave/SurfaceWave flag
        if "body" in os.path.basename(window_param_file):
            wave_type = "#body_wave"
        elif "surface" in os.path.basename(window_param_file):
            wave_type = "#surface_wave"
        else:
            wave_type = ""

        # Output file parameters
        output_file_list.append(os.path.join(cmt_dir, "window_data",
                                             'windows.%03.0f_%03.0f%s.json'
                                             % (lP, hP, wave_type)))

        # Pathfile directory
        yaml_file_path = os.path.join(window_path_dir,
                                      "windows.%03.0f_%03.0f%s.yml"
                                      % (lP, hP, wave_type))

        window_processing_list.append(yaml_file_path)

    # Create dictionary
    logger.info("Windowing Path files:")
    logger.info("--------------------------------")
    logger.info(" ")
    for process_file in window_processing_list:
        logger.info("    " + process_file)
    logger.info(" ")

    return window_processing_list, output_file_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script will create the perturbed Moment tensors in the perturbation
directories.


:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

Last Update: January 2020
"""

from gcmt3d.workflow.generate_sources import write_sources
import argparse


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-f', dest='filename',
                        help='Path to CMTSOLUTION file',
                        required=True, type=str)
    parser.add_argument('-p', dest='param_path',
                        help='Path to param directory',
                        required=True, type=str)
    args = parser.parse_args()

    # Run
    write_sources(args.filename, args.param_path)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

This script writes specfem sources into the respective simulation directories.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""

from gcmt3d.source import CMTSource
from gcmt3d.data import SpecfemSources
from gcmt3d.asdf.utils import smart_read_yaml, is_mpi_env
import os

# Get logger to log progress
from gcmt3d import logger


def write_sources(cmt_filename, param_path):

    # Define parameter directory
    databaseparam_path = os.path.join(param_path,
                                      "Database/DatabaseParameters.yml")
    inversionparam_path = os.path.join(param_path,
                                       "CMTInversion/InversionParams.yml")

    # Load Parameters
    DB_params = smart_read_yaml(databaseparam_path,
                                mpi_mode=is_mpi_env())

    # Inversion Params
    INV_params = smart_read_yaml(inversionparam_path, mpi_mode=is_mpi_env())

    # File and directory
    cmt_dir = os.path.dirname(cmt_filename)
    cmt = CMTSource.from_CMTSOLUTION_file(cmt_filename)
    outdir = os.path.join(cmt_dir, "CMT_SIMs")

    # Basic parameters
    dm = float(INV_params["config"]["dmoment"])      # 10**22 dyne*cm
    dz = float(INV_params["config"]["ddepth"])       # 1000 m
    ddeg = float(INV_params["config"]["dlocation"])  # 0.001 deg

    logger.info(" ")
    logger.info("  Perturbation parameters")
    logger.info("  " + 50 * "*")
    logger.info("  𝚫M: %g" % dm)
    logger.info("  𝚫z: %g" % dz)
    logger.info("  𝚫deg: %g" % ddeg)
    logger.info("  " + 50 * "*" + "\n")

    # Create source creation class
    sfsource = SpecfemSources(cmt, cmt_dir, npar=DB_params['npar'],
                              dm=dm, dx=dz, ddeg=ddeg,
                              verbose=DB_params['verbose'], outdir=outdir)

    # Write sources
    sfsource.write_sources()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from gcmt3d.asdf.convert import ConvertASDF


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='path_file', required=True,
                        help="path file")
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help="verbose flag")
    parser.add_argument('-s', action='store_true', dest='status_bar',
                        help="status bar flag")
    args = parser.parse_args()

    converter = ConvertASDF(args.path_file, args.verbose, args.status_bar)
    converter.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Convert asdf files to sac
"""
import argparse

from gcmt3d.asdf.convert import convert_from_asdf


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', action='store', dest='outputdir', default='.',
                        help="output directory")
    parser.add_argument('filename', help="Input ASDF filename")
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help="verbose")
    parser.add_argument('-s', action='store_true', dest="stationxml",
                        help="Output StationXML files")
    parser.add_argument('-q', action='store_true', dest="quakeml",
                        help="Output Quakeml file")
    args = parser.parse_args()

    convert_from_asdf(
        args.filename, args.outputdir, filetype="sac",
        output_staxml=args.stationxml, output_quakeml=args.quakeml,
        _verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

from gcmt3d.asdf.process import ProcASDF
from gcmt3d.utils.io import smart_read_yaml
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning,
                        module=r'.*numerictypes')
warnings.filterwarnings("ignore", category=UserWarning,
                        module=r'.*asdf_data_set')
warnings.filterwarnings("ignore", category=FutureWarning,
                        module=r'.*numerictypes')


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', action='store', dest='params_file',
                        required=False, help="parameter file", default=None)
    parser.add_argument('-f', action='store', dest='path_file', required=True,
                        help="path file")
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help="verbose flag")
    args = parser.parse_args()

    # Little change to accommodate the full GCMT3D path file.
    if args.params_file is None:
        # Load process path file to get parameter file location
        try:
            params_file = smart_read_yaml(args.path_file)["process_param_file"]
        except KeyError:
            print("The given path file does not contain a parameter file "
                  "destination.")
            return
    else:
        params_file = args.params_file

    proc = ProcASDF(args.path_file, params_file, verbose=args.verbose)
    proc.smart_run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Basic Methods that handles rotation of seismograms as
extension to Obspy. It can rotate `12`, `EN` and `RT`,
forward and backward.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import print_function
from math import cos, sin
from numpy import deg2rad


SMALL_DEGREE = 0.01


def check_orthogonality(azim1, azim2):
    """
    Check if two azimuth are orthogonal, check whether
    (azim1, azim2, vertical) forms a left-hand or right-hand
    coordinate system.
    Remember the defination of azimuth is angle between north
    direction.
    Unit is degree, not radian.
    """
    azim1 = (azim1 + 360) % 360
    azim2 = (azim2 + 360) % 360

    if abs(abs(azim1 - azim2) - 90.0) < SMALL_DEGREE:
        if abs(azim1 - azim2 - 90.0) < SMALL_DEGREE:
            return "right-hand"
        elif abs(azim1 - azim2 + 90.0) < SMALL_DEGREE:
            # should be orthogonal; otherwise return
            return "left-hand"
        else:
            return False
    else:
        # cross 360 degree
        if abs(azim1 - azim2 + 270.0) < SMALL_DEGREE:
            return "right-hand"
        elif abs(azim1 - azim2 - 270.0) < SMALL_DEGREE:
            return "left-hand"
        else:
            return False


def rotate_certain_angle(d1, d2, angle, unit="degree"):
    """
    Basic rotating function which rotate d1 and d2 by angle.
    d1 and d2 should be orthogonal to each other and form a
    'left-handed' coordinate system together with vertical
    component.

    (d1, d2, Vertical) should form a left-handed coordinate system, i.e.,

    ::

        Azimuth_{d2} = Azimuth_{d1} + 90.0 .

    For example, (North, East, Vertical) & (Radial, Transverse, Vertical)
    are both left-handed coordinate systems. The return value (dnew1,
    dnew2, vertic) should also form a left-handed coordinate system.
    The angle is azimuth differnce between d1 and dnew1, i.e.,

    ::

        angle = Azimuth_{dnew1} - Azimuth_{d1}

    :type d1: :class:`~numpy.ndarray`
    :param d1: Data of one of the two horizontal components
    :type d2: :class:`~numpy.ndarray`
    :param d2: Data of one of the two horizontal components
    :type angle: float
    :param angle: component azimuth of data2
    :return: two new components after rotation

    """
    if unit == "degree":
        angle = deg2rad(angle)
    elif unit == "radian":
        angle = angle
    else:
        raise ValueError("Unrecognized unit(%s): 1) degree; 2) radian"
                         % unit)

    if len(d1) != len(d2):
        raise ValueError("Length of d1(%d) and d2(%d) are not the same!"
                         % (len(d1), len(d2)))

    dnew1 = d1 * cos(angle) + d2 * sin(angle)
    dnew2 = -d1 * sin(angle) + d2 * cos(angle)
    return dnew1, dnew2


def rotate_12_rt(d1, d2, baz, azim1, azim2):
    """
    Rotate from any two orthogonal horizontal components to RT components

    :type d1: :class:`~numpy.ndarray`
    :param d1: Data of one of the two horizontal components
    :type d2: :class:`~numpy.ndarray`
    :param d2: Data of the other horizontal components
    :type baz: float
    :param baz: the back azimuth from station to source in degrees
    :type azim1: float
    :param azim1: component azimuth of data1
    :type azim2: float
    :param azim2: component azimuth of data2
    :return: Radial and Transeversal component of seismogram. (None, None)
        returned if input two components are not orthogonal
    """
    status = check_orthogonality(azim1, azim2)
    if not status:
        # raise ValueError("azim1 and azim2 not orthogonal")
        return None, None
    if "right" in status:
        # flip to left-hand
        d1, d2 = d2, d1
        azim1, azim2 = azim2, azim1

    if baz < 0 or baz > 360:
        raise ValueError("Back Azimuth should be between 0 and 360 degree")

    # caculate the angle of rotation
    angle = baz + 180.0 - azim1
    r, t = rotate_certain_angle(d1, d2, angle)

    return r, t


def rotate_rt_12(r, t, baz, azim1, azim2):
    """
    Rotate from any two orthogonal horizontal components to RT components

    :type data1: :class:`~numpy.ndarray`
    :param data1: Data of one of the two horizontal components
    :type data2: :class:`~numpy.ndarray`
    :param data2: Data of the other horizontal components
    :type baz: float
    :param baz: the back azimuth from station to source in degrees
    :type azim1: float
    :param azim1: component azimuth of data1
    :type azim2: float
    :param azim2: component azimuth of data2
    :return: Radial and Transeversal component of seismogram.
    """
    status = check_orthogonality(azim1, azim2)
    if not status:
        raise ValueError("azim1 and azim2 not orthogonal")
    if "left" in status:
        azim = azim1
    elif "right" in status:
        azim = azim2

    if baz < 0 or baz > 360:
        raise ValueError("Back Azimuth should be between 0 and 360 degree")

    # caculate the angle of rotation
    angle = - (baz + 180.0 - azim)
    d1, d2 = rotate_certain_angle(r, t, angle)

    if "right" in status:
        return d2, d1
    elif "left" in status:
        return d1, d2


def rotate_12_ne(d1, d2, azim1, azim2):
    """
    Rotate from any two orthogonal horizontal components to EN components.
    The azimuth of the two horizontal components are specified by azim1
    and azim2.

    :type d1: :class:`~numpy.ndarray`
    :param d1: Data of one of the two horizontal components
    :type d2: :class:`~numpy.ndarray`
    :param d2: Data of the other horizontal components
    :type azim1: float
    :param azim1: component azimuth of data1
    :type azim2: float
    :param azim2: component azimuth of data2
    :return: East and North component of seismogram.
    """
    status = check_orthogonality(azim1, azim2)
    if not status:
        raise ValueError("azim1 and azim2 not orthogonal")
    if "right" in status:
        # flip to left-hand
        d1, d2 = d2, d1
        azim1, azim2 = azim2, azim1

    # caculate the angle of rotation
    n, e = rotate_certain_angle(d1, d2, -azim1)

    return n, e


def rotate_ne_12(n, e, azim1, azim2):
    """
    Rotate from East and North components to give two orghogonal horizontal
    components. Returned values are (d1, d2) and (d1, d2, Vertical) will
    form a left-handed coordinate system.

    :type data1: :class:`~numpy.ndarray`
    :param data1: Data of one of the two horizontal components
    :type data2: :class:`~numpy.ndarray`
    :param data2: Data of the other horizontal components
    :type azim1: float
    :param azim1: component azimuth of data1
    :type azim2: float
    :param azim2: component azimuth of data2
    :return: two horizontal orthogonal seismogram after rotation.
    """
    status = check_orthogonality(azim1, azim2)
    if not status:
        raise ValueError("azim1 and azim2 not orthogonal")
    if "left" in status:
        azim = azim1
    elif "right" in status:
        azim = azim2

    # caculate the angle of rotation
    d1, d2 = rotate_certain_angle(n, e, azim)

    if "right" in status:
        return d2, d1
    elif "left" in status:
        return d1, d2



'''

This script/set of functions is used to create a set of sources for parallel
specfem simulations

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: June 2019

'''

from ...source import CMTSource
from obspy import read_events
import os
import warnings
from copy import deepcopy
import shutil


def replace_file(source, destination):
    """Mini function that replaces a file"""
    if os.path.exists(destination) and os.path.isfile(destination):
        os.remove(destination)
    shutil.copyfile(source, destination)


class SpecfemSources(object):
    '''
    This class handles the writing of specfem sources in form of CMT solutions
    and the stations
    '''

    def __init__(self, cmt, cmt_dir, npar, dm=10.0*24, dx=2., ddeg=0.02,
                 hdur0=False, outdir=None, verbose=False):
        '''
        :param cmt: The original CMT source loaded using CMTSource
        :type cmt: CMTSource
        :param cmt: Directory in which the cmt solution resides within the
                    database
        :type cmt: str
        :param npar: Number of parameters to be inverted for: 6 - only moment
                     tensor, 7 - moment tensor and depth, 9 moment tensor,
                     depth and geolocation.
        :type npar: int
        :param dm: magnitude scalar -- we only need a scalar since the
                   derivative is independent of magnitude
        :type dm: float
        :param dx: depth change constant in m for Frechet derivative
        :type dx: float
        :param ddeg: location change constant for Frechet derivative
        :type ddeg: float
        :param dtshift: change in cmt time shift for Frechet derivative
        :type dtshift: float
        :param hdur0: Sets the half duration to 0 if True.
                      Default False.
        :type hdur0: float
        :param outdir: output directory for sources Should be CMT_SIMs
                       directory as created by the Skeleton class.
        :type outdir: str

        '''

        if type(cmt) != CMTSource:
            raise ValueError('Given CMT parameter not a CMTSource.')
        self.cmt = cmt

        if npar not in [6, 7, 9]:
            raise ValueError('The parameters to be inverted for must be an '
                             + 'integer between 6 and 9.')
        self.npar = npar

        if type(dm) != float:
            raise ValueError('Change in magnitude needs to be a float.')
        self.dm = dm

        if type(dx) != float:
            raise ValueError("Change in depth should be a float")
        self.dx = dx

        if type(ddeg) != float:
            raise ValueError("Change in degrees should be a float")
        self.ddeg = ddeg

        if type(verbose) != bool:
            raise ValueError("Verbose flag must be boolean")
        self.v = verbose

        if outdir is None:
            raise ValueError("The output directory needs wo be set.")
        elif not os.path.exists(outdir):
            os.makedirs(outdir)
            warnings.warn("The chosen output directory does not exist.\n"
                          + "A new one will be created.")

        self.station_dir = os.path.join(cmt_dir, 'station_data')
        self.outdir = outdir

        self.hdur0 = hdur0

    def write_sources(self):
        """Function to write the CMT Solutions to CMT files
        """

        # Write initial CMT solution
        cmtsim_outdir = os.path.join(self.outdir, "CMT",
                                     "DATA")

        # Create new CMT solution
        new_cmt = deepcopy(self.cmt)

        # Set half duration to 0
        if self.hdur0:
            new_cmt.half_duration = 0

        # write file
        new_cmt.write_CMTSOLUTION_file(os.path.join(cmtsim_outdir,
                                                    "CMTSOLUTION"))

        # Print info if verbose flag is True
        if self.v:
            print("%s has been written." % os.path.join(cmtsim_outdir,
                                                        "CMTSOLUTION"))

        # Write the QuakeML to OUTPUT_FILES folder
        src = os.path.join(cmtsim_outdir, "CMTSOLUTION")
        outfiles = os.path.join(self.outdir, "CMT", "OUTPUT_FILES")
        dst = os.path.join(outfiles, "Quake.xml")

        # Write Solution
        self._write_quakeml(src, dst)

        # Print info if verbose flag is True
        if self.v:
            print("%s has been written." % dst)

        # Replace STATIONS FILE
        station_source = os.path.join(self.station_dir, "STATIONS")
        replace_file(station_source, os.path.join(cmtsim_outdir,
                                                  "STATIONS"))

        # Attribute list
        attr = ["m_rr", "m_tt", "m_pp", "m_rt", "m_rp", "m_tp"]

        for index in range(6):

            cmtsim_outdir = os.path.join(self.outdir, "CMT_"
                                         + attr[index][-2:],
                                         "DATA")

            # Create new CMT solution
            new_cmt = deepcopy(self.cmt)

            # make everything zero
            for attribute in attr:
                setattr(new_cmt, attribute, 0)

            # Set half duration to 0
            if self.hdur0:
                new_cmt.half_duration = 0

            # except one variable
            setattr(new_cmt, attr[index], self.dm)

            # write file
            new_cmt.write_CMTSOLUTION_file(os.path.join(cmtsim_outdir,
                                                        "CMTSOLUTION"))

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % os.path.join(cmtsim_outdir,
                                                            "CMTSOLUTION"))

            # Write the QuakeML to OUTPUT_FILES folder
            src = os.path.join(cmtsim_outdir, "CMTSOLUTION")
            outfiles = os.path.join(self.outdir, "CMT_" + attr[index][-2:],
                                    "OUTPUT_FILES")
            dst = os.path.join(outfiles, "Quake.xml")

            self._write_quakeml(src, dst)

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % dst)

            # Replace STATIONS FILE
            station_source = os.path.join(self.station_dir, "STATIONS")
            replace_file(station_source, os.path.join(cmtsim_outdir,
                                                      "STATIONS"))

        if self.npar > 6:

            # Attribute name
            depth_str = "depth_in_m"

            # Set output dir
            cmtsim_outdir = os.path.join(self.outdir, "CMT_depth", "DATA")

            # Create new CMT solution
            new_cmt = deepcopy(self.cmt)

            # change a depth
            setattr(new_cmt, depth_str, new_cmt.depth_in_m + self.dx)

            # Set half duration to 0
            if self.hdur0:
                new_cmt.half_duration = 0

            # write new solution
            new_cmt.write_CMTSOLUTION_file(os.path.join(cmtsim_outdir,
                                                        "CMTSOLUTION"))

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % os.path.join(cmtsim_outdir,
                                                            "CMTSOLUTION"))

            # Write the QuakeML to OUTPUT_FILES folder
            src = os.path.join(cmtsim_outdir, "CMTSOLUTION")
            outfiles = os.path.join(self.outdir, "CMT_depth",
                                    "OUTPUT_FILES")
            dst = os.path.join(outfiles, "Quake.xml")

            self._write_quakeml(src, dst)

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % dst)

            # Replace STATIONS FILE
            station_source = os.path.join(self.station_dir, "STATIONS")
            replace_file(station_source, os.path.join(cmtsim_outdir,
                                                      "STATIONS"))

        if self.npar == 9:

            # Attribute name
            lat_str = "latitude"
            lon_str = "longitude"

            # Create new CMT solution
            new_cmt = deepcopy(self.cmt)

            # change a depth
            setattr(new_cmt, lat_str, new_cmt.latitude + self.ddeg)

            # Set half duration to 0
            if self.hdur0:
                new_cmt.half_duration = 0

            # Set outdir
            cmtsim_outdir = os.path.join(self.outdir, "CMT_lat", "DATA")

            # write new solution
            new_cmt.write_CMTSOLUTION_file(os.path.join(cmtsim_outdir,
                                                        "CMTSOLUTION"))

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % os.path.join(cmtsim_outdir,
                                                            "CMTSOLUTION"))

            # Write the QuakeML to OUTPUT_FILES folder
            src = os.path.join(cmtsim_outdir, "CMTSOLUTION")
            outfiles = os.path.join(self.outdir, "CMT_lat",
                                    "OUTPUT_FILES")
            dst = os.path.join(outfiles, "Quake.xml")

            self._write_quakeml(src, dst)

            # Replace STATIONS FILE
            station_source = os.path.join(self.station_dir, "STATIONS")
            replace_file(station_source, os.path.join(cmtsim_outdir,
                                                      "STATIONS"))

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % dst)

            # Create new CMT solution
            new_cmt = deepcopy(self.cmt)

            # change a depth
            setattr(new_cmt, lon_str, new_cmt.longitude + self.ddeg)

            # Set half duration to 0
            new_cmt.half_duration = 0

            # Set outdir
            cmtsim_outdir = os.path.join(self.outdir, "CMT_lon", "DATA")

            # write new solution
            new_cmt.write_CMTSOLUTION_file(os.path.join(cmtsim_outdir,
                                                        "CMTSOLUTION"))

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % os.path.join(cmtsim_outdir,
                                                            "CMTSOLUTION"))

            # Write the QuakeML to OUTPUT_FILES folder
            src = os.path.join(cmtsim_outdir, "CMTSOLUTION")
            outfiles = os.path.join(self.outdir, "CMT_lon",
                                    "OUTPUT_FILES")
            dst = os.path.join(outfiles, "Quake.xml")

            self._write_quakeml(src, dst)

            # Print info if verbose flag is True
            if self.v:
                print("%s has been written." % dst)

            # Replace STATIONS FILE
            station_source = os.path.join(self.station_dir, "STATIONS")
            replace_file(station_source, os.path.join(cmtsim_outdir,
                                                      "STATIONS"))

    def _write_quakeml(self, source, destination):
        """ Copies CMT solution from source to QuakeML destination."""

        # CMT Source file
        catalog = read_events(source)
        catalog.write(destination, format="QUAKEML")

    def __str__(self):
        string = "-------- CMT Source Writer --------\n"
        string += "Number of parameters to invert for: %d\n" % self.npar
        string += "dM: %f in Nm\n" % self.dm

        if self.npar > 6:
            string += "dx: %f in m\n" % self.dx
        if self.npar >= 9:
            string += "ddeg: %f in degrees\n" % self.ddeg

        return string



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
utils for seismic data download

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from obspy.clients.fdsn import Client
from obspy import UTCDateTime
import os


def read_station_file(station_filename: str):
    stations = []
    with open(station_filename, "rt") as fh:
        for line in fh:
            line_r = line.split()
            # line actually byte type
            # therefore reassignment to string
            stations.append((line_r[1], line_r[0]))
    return stations


def _parse_station_id(station_id: str):
    content = station_id.split("_")
    if len(content) == 2:
        nw, sta = content
        loc = "*"
        comp = "*"
    elif len(content) == 4:
        nw, sta, loc, comp = content
    else:
        raise ValueError("Can't not parse station_id: %s" % station_id)
    return nw, sta, loc, comp


def download_waveform(stations: list, starttime: UTCDateTime,
                      endtime: UTCDateTime, outputdir=None,
                      client=None):
    """
    download wavefrom data from IRIS data center

    :param stations: list of stations, should be list of station ids,
        for example, "II.AAK.00.BHZ". Parts could be replaced by "*",
        for example, "II.AAK.*.BH*"
    """
    if client is None:
        client = Client("IRIS")

    if starttime > endtime:
        raise ValueError("Starttime(%s) is larger than endtime(%s)"
                         % (starttime, endtime))

    if not os.path.exists(outputdir):
        raise ValueError("Outputdir not exists: %s" % outputdir)

    _status = {}
    for station_id in stations:
        error_code = "None"
        network, station, location, channel = _parse_station_id(station_id)

        if outputdir is not None:
            filename = os.path.join(outputdir, "%s.mseed" % station_id)
            if os.path.exists(filename):
                os.remove(filename)
        else:
            filename = None

        try:
            st = client.get_waveforms(
                network=network, station=station, location=location,
                channel=channel, starttime=starttime, endtime=endtime)
            if len(st) == 0:
                error_code = "stream empty"
            if filename is not None and len(st) > 0:
                st.write(filename, format="MSEED")
        except Exception as e:
            error_code = "Failed to download waveform '%s' due to: %s" \
                % (station_id, str(e))
            print(error_code)

        _status[station_id] = error_code

    return {"stream": st, "status": _status}


def download_stationxml(stations: list, starttime: UTCDateTime,
                        endtime: UTCDateTime, outputdir: str = None,
                        client: str = None, level: str = "response"):

    if client is None:
        client = Client("IRIS")

    if starttime > endtime:
        raise ValueError("Starttime(%s) is larger than endtime(%s)"
                         % (starttime, endtime))

    if not os.path.exists(outputdir):
        raise ValueError("Outputdir not exists: %s" % outputdir)

    _status = {}
    refined_stations = []
    for station_id in stations:
        error_code = "None"
        network, station, location, channel = _parse_station_id(station_id)

        if outputdir is not None:
            filename = os.path.join(outputdir, "%s.xml" % station_id)
            if os.path.exists(filename):
                os.remove(filename)
        else:
            filename = None

        try:
            inv = client.get_stations(
                network=network, station=station, location=location,
                channel=channel, starttime=starttime, endtime=endtime,
                level=level)
            if len(inv) == 0:
                error_code = "Inventory Empty"
            if filename is not None and len(inv) > 0:
                inv.write(filename, format="STATIONXML")

            # Add station to refined list to use for waveform download
            refined_stations.append(station_id)

        except Exception as e:
            error_code = "Failed to download StationXML '%s' due to: %s" \
                % (station_id, str(e))
            print(error_code)

        _status[station_id] = error_code

    return inv, _status, refined_stations



"""

This file contains functions and classes to plot stf inversion results.

:copyright:
    Lucas Sawade (lsawade@princeton.edu, 2019)

:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""

import numpy as np
from numpy.fft import fftshift, fftfreq
from .stf import compute_error
import matplotlib.pyplot as plt


def plot_convergence(t, dt, sft, G, u):
    # Compute error
    euc, relerr = compute_error(t, dt, sft, G, u)

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].plot(euc, label="Euclidean norm error")
    ax[0].set_xlabel("Iteration #")
    ax[0].set_ylim([0, np.max(euc)])
    ax[1].plot(relerr, label="Euclidean norm error")
    ax[1].set_xlabel("Iteration #")
    plt.tight_layout()


def plot_stf(t, ft, fwt, iT, fT, skip=5):

    fig, ax = plt.subplots(1, 2, figsize=(7.5, 3.75))

    # STF plotting
    for _i, f in enumerate(ft[::skip]):
        ax[0].plot(t, f, c=(0.25, 0.25, 0.25, 1),
                   alpha=(len(ft[::skip])*0.05+_i)/(len(ft[::skip])*5))
    ax[0].plot(t, ft[-1], c=(0.8, 0.2, 0.2, 1))
    # ax[0].set_xlim([iT, fT])
    ax[0].set_xlabel('Time in [s]')

    # Frequency spectrum
    freq = fftshift(fftfreq(len(fwt[0]), (t[1]-t[0])))
    for _i, fw in enumerate(fwt[::skip]):
        ax[1].plot(freq, np.real(np.abs(fftshift(fw))),
                   c=(0.25, 0.25, 0.25, 1),
                   alpha=(len(fwt[::skip])*0.05+_i)/(len(fwt[::skip])*10))

    ax[1].plot(freq, np.real(np.abs(fftshift(fwt[-1]))), c=(0.8, 0.2, 0.2, 1))
    # ax[1].set_xlim([0, .5*np.max(1/(t[1]-t[0]))])
    # ax[1].set_ylim([0, np.max(fwt)])
    ax[1].set_xlabel('Frequency in [Hz]')
    plt.tight_layout()

    return ax



# -*- coding: utf-8 -*-

import numpy as np

from obspy.geodetics import calc_vincenty_inverse


def get_dist_in_km(station, event, obsd):
    """
    Returns distance in km
    """
    stats = obsd.stats
    station_coor = station.get_coordinates(".".join([stats.network,
                                                     stats.station,
                                                     stats.location,
                                                     stats.channel[:-1]+"Z"]))

    evlat = event.events[0].origins[0].latitude
    evlon = event.events[0].origins[0].longitude

    dist = calc_vincenty_inverse(station_coor["latitude"],
                                 station_coor["longitude"],
                                 evlat, evlon)[0] / 1000

    return dist


def get_time_array(obsd, event):
    stats = obsd.stats
    dt = stats.delta
    npts = stats.npts
    start = stats.starttime - event.events[0].origins[0].time
    return np.arange(start, start+npts*dt, dt)


# raise levels after rayleigh
def generate_user_levels(config, station, event, obsd, synt):
    """Returns a list of acceptance levels
    """
    stats = obsd.stats
    npts = stats.npts

    base_water_level = config.stalta_waterlevel
    base_cc = config.cc_acceptance_level
    base_tshift = config.tshift_acceptance_level
    base_dlna = config.dlna_acceptance_level
    base_s2n = config.s2n_limit

    stalta_waterlevel = np.ones(npts)*base_water_level
    cc = np.ones(npts)*base_cc
    tshift = np.ones(npts)*base_tshift
    dlna = np.ones(npts)*base_dlna
    s2n = np.ones(npts)*base_s2n

    dist = get_dist_in_km(station, event, obsd)

    # Rayleigh
    r_vel = config.min_surface_wave_velocity
    r_time = dist/r_vel

    times = get_time_array(obsd, event)

    for i, time in enumerate(times):
        if time > r_time:
            stalta_waterlevel[i] = base_water_level*2.0
            tshift[i] = base_tshift/3.0
            cc[i] = 0.95
            dlna[i] = base_dlna/3.0
            s2n[i] = 10*base_s2n

    return stalta_waterlevel, tshift, dlna, cc, s2n



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Functions that compare two traces, return the measurements metrics.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
import numpy as np
import matplotlib.pyplot as plt
from obspy import Trace


def least_square_error(data1, data2):
    """
    waveform difference between data1 and data2
    :param data1:
    :param data2:
    :return:
    """
    # least square tests
    err_max = 0.0
    norm = np.linalg.norm
    err = norm(data1 - data2) / np.sqrt(norm(data1) * norm(data2))
    err_max = max(err, err_max)
    return err_max


def cross_correlation(data1, data2):
    """
    :param data1:
    :param data2:
    :return:
    """
    # correlation tests
    corr_min = 1.0
    corr_mat = np.corrcoef(data1, data2)
    corr = np.min(corr_mat)
    corr_min = min(corr, corr_min)
    return corr_min


def trace_length(tr):
    return (tr.stats.npts-1) * tr.stats.delta


def calculate_misfit(_tr1, _tr2, taper_flag=True, taper_percentage=0.05,
                     correlation_flag=True):
    """
    Calculate the misfit between two traces
    :param tr1: trace 1
    :type tr1: Obspy.Trace
    :param tr2: trace 2
    :type tr2: Obspy.Trace
    :param taper_flag: taper the seismogram or not
    :type taper_flag: bool
    :param taper_percentage: the taper percentage
    :type taper_percentage: float
    """
    if not isinstance(_tr1, Trace):
        raise TypeError("Input tr1(type:%s) must be type of obspy.Trace"
                        % type(_tr1))
    if not isinstance(_tr2, Trace):
        raise TypeError("Input tr2(type:%s) must be type of obspy.Trace"
                        % type(_tr2))

    tr1 = _tr1.copy()
    tr2 = _tr2.copy()

    starttime = max(tr1.stats.starttime, tr2.stats.starttime)
    endtime = min(tr1.stats.endtime, tr2.stats.endtime)
    sampling_rate = min(tr1.stats.sampling_rate, tr2.stats.sampling_rate)
    npts = int((endtime - starttime) * sampling_rate)

    tr1.interpolate(sampling_rate, starttime=starttime, npts=npts)
    tr2.interpolate(sampling_rate, starttime=starttime, npts=npts)

    if taper_flag:
        tr1.taper(max_percentage=taper_percentage, type='hann')
        tr2.taper(max_percentage=taper_percentage, type='hann')

    corr_min = cross_correlation(tr1.data, tr2.data)
    err_max = least_square_error(tr1.data, tr2.data)

    # coverage
    tr1_cover = trace_length(tr1) / trace_length(_tr1)
    tr2_cover = trace_length(tr2) / trace_length(_tr2)

    # amplitude diff
    twdiff = [i / sampling_rate for i in range(npts)]
    amp_ref = np.sum(np.abs(tr1.data) + np.abs(tr2.data)) / (2 * npts)
    wdiff = (tr1.data - tr2.data) / amp_ref

    return {"tr1_coverage": tr1_cover, "tr2_coverage": tr2_cover,
            "correlation": corr_min, "error": err_max,
            "time_array": twdiff, "diff_array": wdiff}


def plot_two_trace(tr1, tr2, trace1_tag="trace 1", trace2_tag="trace 2",
                   figname=None):

    if not isinstance(tr1, Trace):
        raise TypeError("Input tr1(type:%s) must be type of obspy.Trace"
                        % type(tr1))
    if not isinstance(tr2, Trace):
        raise TypeError("Input tr2(type:%s) must be type of obspy.Trace"
                        % type(tr2))

    fig = plt.figure(figsize=(20, 10))

    # subplot 1
    plt.subplot(211)
    t1 = tr1.stats.starttime
    t2 = tr2.stats.starttime
    t_ref = max(t1, t2)

    bt = t1 - t_ref
    times1 = [bt + i * tr1.stats.delta for i in range(tr1.stats.npts)]
    plt.plot(times1, tr1.data, linestyle='-', color='r', marker="*",
             markersize=3, label=trace1_tag, markerfacecolor='r',
             markeredgecolor='none')

    bt = t2 - t_ref
    times2 = [bt + i * tr2.stats.delta for i in range(tr2.stats.npts)]
    plt.plot(times2, tr2.data, '-', color="b", linewidth=0.7,
             label=trace2_tag)

    plt.xlim([min(times1[0], times2[0]), max(times1[-1], times2[-1])])
    plt.legend(loc="upper right")

    xmax = plt.xlim()[1]
    ymin = plt.ylim()[0]
    xpos = 0.7 * xmax
    ypos = 0.4 * ymin
    dypos = abs(0.1 * ymin)

    plt.text(xpos, ypos, "trace id:['%s', '%s']" % (tr1.id, tr2.id))
    ypos -= dypos
    plt.text(xpos, ypos, "reference time: %s" % t_ref)
    ypos -= dypos
    plt.text(xpos, ypos, "detaT: [%6.3f, %6.3f]" % (tr1.stats.delta,
                                                    tr2.stats.delta))

    # calcualte misfit
    res = calculate_misfit(tr1, tr2)

    ypos -= dypos
    plt.text(xpos, ypos, "coverage:[%6.2f%% %6.2f%%]"
             % (res["tr1_coverage"] * 100, res["tr2_coverage"] * 100))
    ypos -= dypos
    plt.text(xpos, ypos, "min correlation: % 6.4f" % res["correlation"])
    ypos -= dypos
    plt.text(xpos, ypos, "max error: % 6.4f" % res["error"])
    plt.grid()

    # subplot 2
    plt.subplot(212)
    plt.plot(res["time_array"], res["diff_array"], 'g',
             label="amplitude difference")
    plt.legend()
    plt.xlim([min(times1[0], times2[0]), max(times1[-1], times2[-1])])
    plt.grid()

    plt.tight_layout()
    if figname is None:
        plt.show()
    else:
        plt.savefig(figname)

    plt.close(fig)



"""
This file contains the class to run specfem for the database.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: May 2019


"""
import os
import re
import subprocess
import shutil
from shlex import split


class RunSimulation(object):
    """Class handles the running of specfem after its directory has been
    added to the database. """

    def __init__(self, earthquake_dir, N=1, npn=24, n=24, memory_req=None,
                 npar=9, walltime="00:30:00", modules=None,
                 gpu_module=None, GPU_MODE=False, verbose=False):
        """
        Initializes Run parameters

        Args:
            database_Cdir: string with directory name
            N: integer with number of Nodes
            n: integer with number of tasks
            npn: integer with tasks per node
            npar: integer number of parameters
            walltime: string with max time "hh:mm:ss"
            modules: list of modules to be loaded for the run (for comilers
                     etc.) Default ['intel', 'openmpi']
            gpu_module: GPU module to be loaded. Default 'cudatoolkit'.
            verbose: boolean deciding on whether to print stuff

        Returns: Nothing really, it just runs specfem with the above options

        """

        self.earthquake = earthquake_dir
        self.simdir = os.path.join(self.earthquake, "CMT_SIMs")
        self.N = N
        self.n = n
        self.npn = npn
        self.memory_req = memory_req
        self.walltime = walltime
        self.modules = modules
        self.gpu_module = gpu_module
        self.GPU_MODE = GPU_MODE
        self.attr = ["CMT", "CMT_rr", "CMT_tt", "CMT_pp", "CMT_rt", "CMT_rp",
                     "CMT_tp", "CMT_depth", "CMT_lat", "CMT_lon"]
        if npar in [6, 7, 9]:
            self.npar = npar
        else:
            raise ValueError("Wrong number. must be 6, 7, or 9.")
        self.batchdir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "batch")
        self.v = verbose

    def __call__(self):
        """Runs the Simulation using the shell and batch files in the batch
        subdirectory."""

        if self.GPU_MODE is True and type(self.GPU_MODE) is bool:

            if self.v:
                print("GPU mode!")

            # batch driver wrapper
            batchwrapper = os.path.join(self.batchdir, "drive_GPU.sh")

            # batch driver script
            batchscript = os.path.join(self.batchdir, "drive_GPU.sbatch")

            # Create command
            bashCommand = "%s %s %s %s %s %s %s %s '%s' %s %s %s" \
                          % (batchwrapper,
                             self.N,
                             self.n,
                             self.npn,
                             self.memory_req,
                             self.npar,
                             self.simdir,
                             self.walltime,
                             " ".join(self.modules),
                             self.gpu_module,
                             int(self.v),
                             batchscript)

            print(bashCommand)

        else:

            if self.v:
                print("CPU mode!")

            # batch driver wrapper
            batchwrapper = os.path.join(self.batchdir, "drive.sh")

            # batch driver script
            batchscript = os.path.join(self.batchdir, "drive.sbatch")

            # Create command
            bashCommand = "%s %s %s %s %s %s '%s' %s %s" % (batchwrapper,
                                                            self.N,
                                                            self.n,
                                                            self.npar,
                                                            self.simdir,
                                                            self.walltime,
                                                            " ".join(
                                                                self.modules),
                                                            int(self.v),
                                                            batchscript)

        # Send command
        process = subprocess.run(split(bashCommand), check=True, text=True,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

        # catch outputs
        if self.v:
            print(bashCommand)
            print("Command has been sent.")
            print("Output:\n", process.stdout)
            print("Errors:\n", process.stderr)

    def replace_STATIONS(self, statfile):
        """This function handles the replacement of the STATION file in the
        database directory."""

        if (self.npar is None) or (self.simdir is None):
            raise ValueError("No number of parameters or Sim dir given")
        else:
            for at in self.attr[:self.npar+1]:
                newstatfile = os.path.join(self.simdir, at, "DATA", "STATIONS")
                self._replace_file(statfile, newstatfile)

    def clean_up(self):
        """This function cleans out the simulation directory, getting rid of
        unnecessary MPI and database files. The parameter files are first
        copied to the directory OUTPUT_FILES and then `bin`, `DATA` and
        `DATABASES_MPI` are deleted since they are irrelevant for future
        reproducibility."""

        for at in self.attr[:self.npar+1]:
            # Copy Par_file CMTSOLUTION and STATIONS files to the OUTPUTFILES
            # directory
            atsimdir = os.path.join(self.simdir, at)
            datadir = os.path.join(atsimdir, "DATA")
            outdir = os.path.join(atsimdir, "OUTPUT_FILES")

            # Stations
            self._replace_file(os.path.join(datadir, "STATIONS"),
                               os.path.join(outdir, "STATIONS"))
            # Par_file
            self._replace_file(os.path.join(datadir, "Par_file"),
                               os.path.join(outdir, "Par_file"))

            # CMTSOLUTION
            self._replace_file(os.path.join(datadir, "CMTSOLUTION"),
                               os.path.join(outdir, "CMTSOLUTION"))

            # Remove bin symlink
            self._remove_file(os.path.join(atsimdir, "bin"))

            # remove unnecessary
            self._remove_dir(os.path.join(atsimdir, "DATA"))
            self._remove_dir(os.path.join(atsimdir, "DATABASES_MPI"))

    def _replace_file(self, source, destination):
        """Mini function that replaces a directory"""
        if os.path.exists(source):
            if os.path.exists(destination) and os.path.isfile(destination):
                if self.v:
                    print("File already here. Removing %s ..." % destination)
                os.remove(destination)

            if self.v:
                print("Copying %s to %s ..." % (source, destination))
            shutil.copyfile(source, destination)
        else:
            print("%s does not exist. Not copied to %s."
                  % (source, destination))

    def _remove_file(self, file):
        """ Mini function to remove file."""
        if os.path.exists(file) and os.path.isfile(file):
            if self.v:
                print("Removing %s ..." % file)
            os.remove(file)
        else:
            print("%s does not exist. Hence not deleted." % file)

    def _remove_dir(self, direc):
        """ Mini function to remove directory."""
        if os.path.exists(direc) and os.path.isdir(direc):
            if self.v:
                print("Removing %s ..." % direc)
            shutil.rmtree(direc)
        else:
            print("%s does not exist. Hence not deleted." % direc)

    def __str__(self):
        """string return"""
        string = ""
        string += "Earthquake directory: %s\n" % self.earthquake
        string += "Simulation directory: %s\n" % self.simdir
        string += "Number of Nodes: %d\n" % self.N
        string += "Number of Tasks: %d\n" % self.n
        string += "Number of Parameters: %d\n\n" % self.npar
        return string


class DATAFixer(object):
    """Not necessary but it handles the fixing of the parfile"""

    def __init__(self, specfemdir,
                 nodes=1,
                 tasks=24,
                 tasks_per_node=24,
                 memory_req=40000,
                 walltime="00:30:00",
                 walltime_solver=None,
                 cc="icc",
                 cpp="icpc",
                 mpicc="mpicc",
                 f90="ifort",
                 mpif90="mpifort",
                 modules=['intel', 'openmpi'],
                 gpu_module=None,
                 gpu_version=None,
                 NEX_XI=128, NEX_ETA=128,
                 NPROC_XI=1, NPROC_ETA=1,
                 GPU_MODE=False,
                 GPU_RUNTIME=1,
                 GPU_PLATFORM='NVIDIA',
                 GPU_DEVICE='Tesla',
                 ADIOS_ENABLED=False,
                 ADIOS_FOR_FORWARD_ARRAYS=True,
                 ADIOS_FOR_MPI_ARRAYS=True,
                 ADIOS_FOR_ARRAYS_SOLVER=True,
                 ADIOS_FOR_SOLVER_MESHFILES=True,
                 ADIOS_FOR_AVS_DX=True,
                 ADIOS_FOR_KERNELS=True,
                 ADIOS_FOR_MODELS=True,
                 ADIOS_FOR_UNDO_ATTENUATION=True,
                 ROTATE_SEISMOGRAMS_RT=True,
                 RECORD_LENGTH=10,
                 MODEL="s40rts",
                 WRITE_SEISMOGRAMS_BY_MASTER=True,
                 OUTPUT_SEISMOS_ASCII_TEXT=False,
                 OUTPUT_SEISMOS_SAC_ALPHANUM=False,
                 OUTPUT_SEISMOS_SAC_BINARY=False,
                 OUTPUT_SEISMOS_ASDF=False,
                 MOVIE_SURFACE=False,
                 MOVIE_VOLUME=False,
                 MOVIE_COARSE=False,
                 verbose=False):
        """
        Initializes Run parameters in the Parfile

        :param specfemdir: string with directory name
        :param nodes: number of computational nodes for the mesher. Default `1`
        :param tasks: number of tasks per node for the mesher. Default `24`
        :param tasks_per_node: maximum number of tasks per node. Default `24`.
        :param walltime: Requested server time for the nodes.
                         Default `00:30:00`
        :param walltime_solver: walltime of mesher and solver can be different.
        f90: string with fortran compiler
        :param mpif90: string with fortran mpi compiler
        :param cc: string with c compiler
        :param cpp: string with c++ compiler
        :param mpicc: string with c mpi compiler
        :param modules: list of modules to be loaded for the run (for comilers
                 etc.) Default ['intel', 'openmpi']
        :param gpu_module: GPU module to be loaded. Default 'None'.
        :param gpu_version: version that is used to compile specfem.
        :param walltime: Requested server time for the nodes. Default `None`.
                  `None` since this class is used for fixing the `Par_file`
        :param NEX_XI: Number of elements along the chunk (s. Specfem Manual).
                Default `128`.
        :param NEX_ETA: Number of elements along the first chunk (s. Specfem
                 Manual). Default `128`.
        :param NPROC_XI: Number of MPI processors (s. Specfem Manual).
                  Default `1`.
        :param NPROC_ETA: Number of MPI processors (s. Specfem Manual).
                   Default `1`.
        :param GPU_MODE: Set whether specfem should be GPU enabled (not yet
                supported). Default `False`.
        :param ADIOS_ENABLED: Set whether ADIOS should be turned on (not yet
                       supported). Default `False`
        :param RECORD_LENGTH: length of the final seismic record in minutes.
                       Default `1`.
        :param MODEL: velocity model, right now only 3D models are supported.
               Default `'s40rts'`
        :param ROTATE_SEISMOGRAMS_RT: This sets whether the seismograms are
                               output as as NEZ or RTZ if True RTZ is
                               chosen. Default `True`.
        :param WRITE_SEISMOGRAMS_BY_MASTER: Write seismograms by master job.
                                     Default `False`.
        :param OUTPUT_SEISMOS_ASCII_TEXT: Output seismograms in `ASCII` format.
                                   Default `False`.
        :param OUTPUT_SEISMOS_SAC_ALPHANUM: Output seismograms in `ASCII`
                                            format. Default `False`.
        :param OUTPUT_SEISMOS_SAC_BINARY: Output seismograms in `ASCII` format.
                                   Default `False`.
        :param OUTPUT_SEISMOS_ASDF: Output seismograms in `ASCII` format.
                             Default `False`.
        :param MOVIE_SURFACE: Write movie on the surface. Default `False`
        :param MOVIE_VOLUME: Write movie within the volume. Default `False`
        :param MOVIE_COARSE: Write coarse movie. Default `False`
        :param verbose: boolean deciding on whether to print stuff.
                Default `False`.

        Returns: Nothing really it just runs specfem with the above options
        """

        # Specfem parameters
        self.specfemdir = specfemdir

        # Slurm Resources for mesher
        self.nodes = nodes
        self.tasks = tasks
        self.tasks_per_node = tasks_per_node
        self.memory_req = memory_req
        self.walltime = walltime  # walltime for mesher

        # The data fixer can also run specfem for that a specific walltime
        # has to be set
        self.walltime_solver = walltime_solver

        # Compilers
        self.mpif90 = mpif90
        self.f90 = f90
        self.cc = cc
        self.cpp = cpp
        self.mpicc = mpicc

        # Modules
        self.modules = modules
        self.gpu_module = gpu_module

        # MPI STUFF
        self.NPROC_XI = NPROC_XI
        self.NPROC_ETA = NPROC_ETA
        self.NEX_XI = NEX_XI
        self.NEX_ETA = NEX_ETA

        # GPU acceleration parameters
        self.GPU_MODE = GPU_MODE
        self.GPU_RUNTIME = GPU_RUNTIME
        self.GPU_PLATFORM = GPU_PLATFORM
        self.GPU_DEVICE = GPU_DEVICE
        self.gpu_version = gpu_version

        # ADIOS parameters
        self.ADIOS_ENABLED = ADIOS_ENABLED
        self.ADIOS_FOR_FORWARD_ARRAYS = ADIOS_FOR_FORWARD_ARRAYS
        self.ADIOS_FOR_MPI_ARRAYS = ADIOS_FOR_MPI_ARRAYS
        self.ADIOS_FOR_ARRAYS_SOLVER = ADIOS_FOR_ARRAYS_SOLVER
        self.ADIOS_FOR_SOLVER_MESHFILES = ADIOS_FOR_SOLVER_MESHFILES
        self.ADIOS_FOR_AVS_DX = ADIOS_FOR_AVS_DX
        self.ADIOS_FOR_KERNELS = ADIOS_FOR_KERNELS
        self.ADIOS_FOR_MODELS = ADIOS_FOR_MODELS
        self.ADIOS_FOR_UNDO_ATTENUATION = ADIOS_FOR_UNDO_ATTENUATION

        # PARAMETERS
        self.vmodel = MODEL
        self.RECORD_LENGTH = RECORD_LENGTH

        # Processing
        self.ROTATE_SEISMOGRAMS_RT = ROTATE_SEISMOGRAMS_RT

        # OUTPUT seismograms as:
        self.WRITE_SEISMOGRAMS_BY_MASTER = WRITE_SEISMOGRAMS_BY_MASTER
        self.OUTPUT_SEISMOS_ASCII_TEXT = OUTPUT_SEISMOS_ASCII_TEXT
        self.OUTPUT_SEISMOS_SAC_ALPHANUM = OUTPUT_SEISMOS_SAC_ALPHANUM
        self.OUTPUT_SEISMOS_SAC_BINARY = OUTPUT_SEISMOS_SAC_BINARY
        self.OUTPUT_SEISMOS_ASDF = OUTPUT_SEISMOS_ASDF

        # Visualization
        self.MOVIE_SURFACE = MOVIE_SURFACE
        self.MOVIE_VOLUME = MOVIE_VOLUME
        self.MOVIE_COARSE = MOVIE_COARSE

        self.v = verbose

        # Set the directory for the batch script
        self.batchdir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "batch")

    def fix_parfiles(self):
        """This function changes the number of nodes within the parfile
        in each subdirectory of the CMT_SIMs directory."""

        parfile = os.path.join(self.specfemdir, "DATA/Par_file")

        # Make sure it's a global simulation:
        self.replace_varval(parfile, "NCHUNKS", str(6))

        # Replace elements along surface of the two sides of first chunk
        self.replace_varval(parfile, "NEX_XI", self.NEX_XI)
        self.replace_varval(parfile, "NEX_ETA", self.NEX_ETA)

        # Replace number of MPI processors along the two sides of the first
        # chunk
        self.replace_varval(parfile, "NPROC_XI", self.NPROC_XI)
        self.replace_varval(parfile, "NPROC_ETA", self.NPROC_ETA)

        # Check whether velocity model has to change
        if self.vmodel not in [
                "transversely_isotropic_prem_plus_3D_crust_2.0",
                "3D_anisotropic", "3D_attenuation", "s20rts",
                "s40rts", "s362ani", "s362iso", "s362wmani",
                "s362ani_prem", "s362ani_3DQ", "s362iso_3DQ",
                "s29ea", "sea99_jp3d1994", "sea99", "jp3d1994",
                "heterogen", "full_sh", "sgloberani_aniso",
                "sgloberani_iso"]:
            raise ValueError("Wrong velocity model name. Existing model is"
                             " used.")
        else:
            self.replace_varval(parfile, "MODEL", self.vmodel)

        # Replace RECORD_LENGTH_IN_MINUTES
        self.replace_varval(parfile, "RECORD_LENGTH_IN_MINUTES", "%s.0d0" %
                            self.RECORD_LENGTH)

        # Processing
        self.replace_varval(parfile, "ROTATE_SEISMOGRAMS_RT",
                            self.ROTATE_SEISMOGRAMS_RT)

        # OUTPUT seismograms as:
        self.replace_varval(parfile, "WRITE_SEISMOGRAMS_BY_MASTER",
                            self.WRITE_SEISMOGRAMS_BY_MASTER)
        self.replace_varval(parfile, "OUTPUT_SEISMOS_ASCII_TEXT",
                            self.OUTPUT_SEISMOS_ASCII_TEXT)
        self.replace_varval(parfile, "OUTPUT_SEISMOS_SAC_ALPHANUM",
                            self.OUTPUT_SEISMOS_SAC_ALPHANUM)
        self.replace_varval(parfile, "OUTPUT_SEISMOS_SAC_BINARY",
                            self.OUTPUT_SEISMOS_SAC_BINARY)
        self.replace_varval(parfile, "OUTPUT_SEISMOS_ASDF",
                            self.OUTPUT_SEISMOS_ASDF)

        # Visualization
        self.replace_varval(parfile, "MOVIE_SURFACE", self.MOVIE_SURFACE)
        self.replace_varval(parfile, "MOVIE_VOLUME", self.MOVIE_VOLUME)
        self.replace_varval(parfile, "MOVIE_COARSE", self.MOVIE_COARSE)

        # Acceleration parameters supported
        self.replace_varval(parfile, "GPU_MODE", self.GPU_MODE)
        self.replace_varval(parfile, "GPU_RUNTIME", self.GPU_RUNTIME)
        self.replace_varval(parfile, "GPU_PLATFORM", self.GPU_PLATFORM)
        self.replace_varval(parfile, "GPU_DEVICE", self.GPU_DEVICE)

        # ADIOS
        self.replace_varval(parfile, "ADIOS_ENABLED", self.ADIOS_ENABLED)
        self.replace_varval(parfile, "ADIOS_FOR_MPI_ARRAYS",
                            self.ADIOS_FOR_MPI_ARRAYS)
        self.replace_varval(parfile, "ADIOS_FOR_ARRAYS_SOLVER",
                            self.ADIOS_FOR_ARRAYS_SOLVER)
        self.replace_varval(parfile, "ADIOS_FOR_SOLVER_MESHFILES",
                            self.ADIOS_FOR_SOLVER_MESHFILES)
        self.replace_varval(parfile, "ADIOS_FOR_AVS_DX", self.ADIOS_FOR_AVS_DX)
        self.replace_varval(parfile, "ADIOS_FOR_KERNELS",
                            self.ADIOS_FOR_KERNELS)
        self.replace_varval(parfile, "ADIOS_FOR_MODELS", self.ADIOS_FOR_MODELS)
        self.replace_varval(parfile, "ADIOS_FOR_UNDO_ATTENUATION",
                            self.ADIOS_FOR_UNDO_ATTENUATION)

    def configure_and_make(self):
        """ Configures specfem and runs 'make all' in the specfemdirectory."""

        if self.GPU_MODE and type(self.GPU_MODE) is bool:

            if self.v:
                print("GPU mode!")

            # Create configuration command
            configure_script = os.path.join(self.batchdir, "configure_GPU.sh")

            # Create command compilation command
            bashCommand = "%s %s %s %s %s %s %s '%s' %s %s" \
                          % (configure_script,
                             self.specfemdir,
                             self.mpif90,
                             self.f90,
                             self.cc,
                             self.cpp,
                             self.mpicc,
                             " ".join(self.modules),
                             self.gpu_module,
                             self.gpu_version)

        else:
            if self.v:
                print("CPU mode!")

            # Create configuration command
            configure_script = os.path.join(self.batchdir, "configure.sh")

            # Create command compilation command
            bashCommand = "%s %s %s %s %s %s %s '%s'" \
                          % (configure_script,
                             self.specfemdir,
                             self.mpif90,
                             self.f90,
                             self.cc,
                             self.cpp,
                             self.mpicc,
                             " ".join(self.modules))

        if self.v:
            # Send command
            process = subprocess.run(split(bashCommand), check=True, text=True)
        else:
            # Send command
            process = subprocess.run(split(bashCommand), check=True, text=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)

        # catch outputs
        if self.v:
            print(bashCommand)
            print("Command has been sent.")
            print("Output:\n", process.stdout)
            print("Errors:\n", process.stderr)
            print("Compilation DONE!\n")

    def run_mesher(self):
        """This function runs the mesher after """

        if self.GPU_MODE and type(self.GPU_MODE) is bool:

            if self.v:
                print("GPU mode!")

            # batch driver script
            batchscript = os.path.join(self.batchdir, "mesh_GPU.sbatch")

            # Create command -N nodes, -n tasks, -D change directory
            bashCommand = "sbatch -N %s -n %s --ntasks-per-node %s -D %s -t " \
                          "%s --gres=gpu:%s -p pReserved %s" \
                          % (self.nodes,
                             self.tasks,
                             self.tasks_per_node,
                             self.specfemdir,
                             self.walltime,
                             self.tasks_per_node,
                             batchscript)

        else:

            if self.v:
                print("CPU mode!")

            # batch driver script
            batchscript = os.path.join(self.batchdir, "mesh.sbatch")

            # Create command -N nodes, -n tasks, -D change directory
            bashCommand = "sbatch -N %s -n %s -D %s -t %s %s" \
                          % (self.nodes,
                             self.tasks,
                             self.specfemdir,
                             self.walltime,
                             batchscript)

        if self.v:
            # Send command
            process = subprocess.run(split(bashCommand), check=True, text=True)
        else:
            # Send command
            process = subprocess.run(split(bashCommand), check=True, text=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)

        # catch outputs
        if self.v:
            print(bashCommand)
            print("Command has been sent.")
            print("Output:\n", process.stdout)
            print("Errors:\n", process.stderr)
            print("Meshing SUBMITTED!\n")

    def run_solver(self):
        """This function runs the solver within the"""

        # Check of solver runtime is set
        if self.GPU_MODE and type(self.GPU_MODE) is bool:

            if self.v:
                print("GPU mode!")

            # batch driver script
            batchscript = os.path.join(self.batchdir, "solver_GPU.sbatch")

            # Create command -N nodes, -n tasks, -D change directory
            bashCommand = "sbatch -N %s -n %s --ntasks-per-node %s -D %s " \
                          "-t %s --mem %s --gres=gpu:%s -p pReserved %s" \
                          % (self.nodes,
                             self.tasks,
                             self.tasks_per_node,
                             self.specfemdir,
                             self.walltime_solver,
                             self.memory_req,
                             self.tasks_per_node,
                             batchscript)

        else:
            # batch driver script
            batchscript = os.path.join(self.batchdir, "solver.sbatch")

            # Create command -N nodes, -n tasks, -D change directory
            bashCommand = "sbatch -N %s -n %s -D %s -t %s %s" \
                          % (self.nodes,
                             self.tasks,
                             self.specfemdir,
                             self.walltime_solver,
                             batchscript)

        if self.v:
            # Send command
            process = subprocess.run(split(bashCommand), check=True, text=True)
        else:
            # Send command
            process = subprocess.run(split(bashCommand), check=True, text=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)

        # catch outputs
        if self.v:
            print(bashCommand)
            print("Command has been sent.")
            print("Output:\n", process.stdout)
            print("Errors:\n", process.stderr)

    @staticmethod
    def replace_varval(filename, var, newval):
        """ This function updates the value of a function within a text file

        !Note! That this file is customized to change the specfem `Par_file`,
               Meaning, if newval is `True` or `False` it will be replaced with
               `.true.` or `.false` respectively.

        Args:
            var: variable name -- string
            newval: new variable value, string, number or list

        Throws and error if variable doesnt exist or has multiple definitions.
        """

        file = open(filename, 'r+')
        content_lines = []
        counter = 0

        # Specfem Par_file modification:
        if type(newval) == bool:
            if newval is True:
                newval = '.true.'
            elif newval is False:
                newval = '.false.'

        for line in file:
            # more robust than simple string comparison
            if re.match(" *" + var + " *=", line):
                counter += 1
                # Split the line into to at the equal sign
                line_components = line.split('=')

                # set the value of the line again
                line_components[1] = str(newval) + "\n"
                updated_line = "= ".join(line_components)
                content_lines.append(updated_line)
            else:
                content_lines.append(line)

        # Check whether variable is in file or has multiple definitions
        if counter == 0:
            raise ValueError("Variable not in file.")
        elif counter > 1:
            raise ValueError("Variable is defined in multiple places. Cannot "
                             "overwrite.")
        else:
            file.seek(0)
            file.truncate()
            file.writelines(content_lines)

        file.close()

    @staticmethod
    def get_val(filename, var):
        """ Function searches file for variable and returns that value as a
        string.

        Args:
            filename: string
            var: string

        Returns:
            val
        """

        file = open(filename, 'r+')
        counter = 0

        for line in file:
            # more robust than simple string comparison
            if re.match(" *" + var + ' *=', line):
                counter += 1
                # Split the line into to at the equal sign
                line_components = line.split('=')

                # set the value of the line again
                val = line_components[1]

        # Check whether variable is in file or has multiple definitions
        if counter == 0:
            raise ValueError("Variable not in file.")
        elif counter > 1:
            raise ValueError("Variable is defined in multiple places. Cannot "
                             "overwrite.")
        else:
            return val



"""

This script contains functions to write the inversion dictionaries to the
filesystem.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: June 2019

"""

from gcmt3d.data.management.inversion_dicts \
    import create_full_inversion_dict_list
from gcmt3d.data.management.inversion_dicts \
    import create_g3d_inversion_dict_list
from gcmt3d.data.management.inversion_dicts import write_inversion_dicts
import yaml
import os


def read_yaml_file(filename):
    """read yaml file"""
    with open(filename) as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


def inversion_dictionaries(cmt_file_db, param_path):
    """ This function writes the inversion dictionaries

    :param cmt_file_db:
    :return:
    """

    # Processing param dir
    process_obs_dir = os.path.join(param_path, "ProcessObserved")
    process_syn_dir = os.path.join(param_path, "ProcessSynthetic")

    # Window param dir
    window_process_dir = os.path.join(param_path, "CreateWindows")

    # Create inversion dictionaries
    inv_dict_list, filename_list = \
        create_full_inversion_dict_list(cmt_file_db,
                                        process_obs_dir,
                                        process_syn_dir,
                                        window_process_dir)

    # Write the dictionaries
    write_inversion_dicts(inv_dict_list, filename_list)


def grid_search_dictionaries(cmt_file_db, param_path):
    """ This function writes the inversion dictionaries

    :param cmt_file_db:
    :return:
    """

    # Processing param dir
    process_obs_dir = os.path.join(param_path, "ProcessObserved")
    process_syn_dir = os.path.join(param_path, "ProcessSynthetic")

    # Window param dir
    window_process_dir = os.path.join(param_path, "CreateWindows")

    # Create inversion dictionaries
    inv_dict_list, filename_list = \
        create_g3d_inversion_dict_list(cmt_file_db,
                                       process_obs_dir,
                                       process_syn_dir,
                                       window_process_dir)

    # Write the dictionaries
    write_inversion_dicts(inv_dict_list, filename_list)



from .asdf.utils import smart_read_yaml, is_mpi_env  # NOQA
from .data.management.skeleton import DataBaseSkeleton  # NOQA
import logging
from .log_util import modify_logger

# Setup the logger
logger = logging.getLogger("gcmt3d")

# Sets up the formatting and everything..
modify_logger(logger)

'''The different levels are

 0: logging.NOTSET
10: logging.DEBUG
15: logging.VERBOSE
20: logging.INFO
30: logging.WARNING
40: logging.ERROR
50: logging.CRITICAL

'''
logger.setLevel(logging.VERBOSE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Methods that handles signal data processing for stream(Obspy.Stream)

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""

from __future__ import (division, print_function, absolute_import)
from obspy.signal.invsim import cosine_sac_taper
from obspy.signal.util import _npts2nfft
from obspy import Stream, Trace
import numpy as np
from .rotate import rotate_stream


def check_array_order(array, order="ascending"):
    """
    Check whether a array is in ascending order or descending order
    :param array: the input array
    :param order: "ascending" or "descending"
    :return:
    """
    array = np.array(array)
    if order not in ("descending", "ascending"):
        raise ValueError("Order should be either ascending or descending")

    if order == "descending":
        array *= -1

    return (array == sorted(array)).all()


def flex_cut_trace(trace, cut_starttime, cut_endtime, dynamic_npts=0):
    """
    not cut strictly(but also based on the original trace length)

    :param trace: input trace
    :type trace: obspy.Trace
    :param cut_starttime: starttime of cutting
    :type cut_starttime: obspy.UTCDateTime
    :param cut_endtime: endtime of cutting
    :type cut_endtime: obspy.UTCDateTime
    """
    if not isinstance(trace, Trace):
        raise TypeError("flex_cut_trace method only accepts obspy.Trace "
                        "as the first argument")

    delta = trace.stats.delta
    cut_starttime = cut_starttime - dynamic_npts * delta
    cut_endtime = cut_endtime + dynamic_npts * delta
    trace.trim(cut_starttime, cut_endtime)


def flex_cut_stream(st, cut_start, cut_end, dynamic_npts=0):
    """
    Flexible cut stream. But checks for the time.

    :param st: input stream
    :param cut_start: cut starttime
    :param cut_end: cut endtime
    :param dynamic_npts: the dynamic number of points before cut_start
        and after
        cut_end
    :return: the cutted stream
    """
    if not isinstance(st, Stream):
        raise TypeError("flex_cut_stream method only accepts obspy.Stream "
                        "the first Argument")
    new_st = Stream()
    count = 0
    for tr in st:
        flex_cut_trace(tr, cut_start, cut_end, dynamic_npts=dynamic_npts)
        # throw out small piece of data at this step
        if tr.stats.starttime <= cut_start and tr.stats.endtime >= cut_end:
            new_st.append(tr)
            count += 1
    if count == 0:
        raise ValueError("None of traces in Stream satisfy the "
                         "cut time length")
    return new_st


def filter_stream(st, pre_filt):
    """
    Filter a stream

    :param st:
    :param per_filt:
    :return:
    """
    if not isinstance(st, Stream):
        raise TypeError("Input st should be type of Stream")
    for tr in st:
        filter_trace(tr, pre_filt)


def filter_trace(tr, pre_filt):
    """
    Perform a frequency domain taper mimicing the behavior during the
    response removal, without a actual response removal.

    :param tr: input trace
    :param pre_filt: frequency array(Hz) in ascending order, to define
        the four corners of filter, for example, [0.01, 0.1, 0.2, 0.5].
    :type pre_filt: Numpy.array or list
    :return: filtered trace
    """
    if not isinstance(tr, Trace):
        raise TypeError("First Argument should be trace: %s" % type(tr))
    if len(pre_filt) != 4:
        raise ValueError("Length of filter must be 4(corner frequencies)")
    if not check_array_order(pre_filt, order="ascending"):
        raise ValueError("Frequency band should be in ascending order: %s"
                         % pre_filt)

    data = tr.data.astype(np.float64)
    origin_len = len(data)
    if origin_len == 0:
        return

    # smart calculation of nfft dodging large primes
    nfft = _npts2nfft(len(data))

    fy = 1.0 / (tr.stats.delta * 2.0)
    freqs = np.linspace(0, fy, nfft // 2 + 1)

    # Transform data to Frequency domain
    data = np.fft.rfft(data, n=nfft)
    data *= cosine_sac_taper(freqs, flimit=pre_filt)
    data[-1] = abs(data[-1]) + 0.0j
    # transform data back into the time domain
    data = np.fft.irfft(data)[0:origin_len]
    # assign processed data and store processing information
    tr.data = data


def interpolate_stream(stream, sampling_rate, starttime=None, npts=None):
    """
    For a fairly large stream, use stream.interpolate() is not a wise
    choice since if there is one trace fails, then the whole interpolation
    will stop. So it is better to operate interpolation on the trace
    level
    """
    st_new = Stream()
    if not isinstance(stream, Stream):
        raise TypeError("Input stream must be type of obspy.Stream")
    for tr in stream:
        try:
            tr.interpolate(sampling_rate, starttime=starttime, npts=npts)
            st_new.append(tr)
        except ValueError as err:
            print("Error in interpolation on '%s':%s" % (tr.id, err))
    return st_new


def process_stream(st, inventory=None, remove_response_flag=False,
                   water_level=60, filter_flag=False, pre_filt=None,
                   starttime=None, endtime=None,
                   resample_flag=False, sampling_rate=1.0,
                   taper_type="hann", taper_percentage=0.05,
                   rotate_flag=False, event_latitude=None,
                   event_longitude=None, sanity_check=False):
    """
    Stream processing function defined for general purpose of tomography.
    The advantage of using Stream, rather than than Trace, is that rotation
    could be operated if the Stream contains multiple channels. But this
    function also deals with Trace, but you need to set rotate_flag to
    False

    :param st: input stream
    :type st: obspy.Stream
    :param remove_response_flag: flag for remove instrument response. If True,
        then inv should be specified, and filter_flag would not be taken caren
        of. If you want just filter the seismogram, please leave this to False
        and set filter_flag to True.
    :type remove_response_flag: bool
    :param inventory: station inventory information
    :type inventory: obspy.Inventory
    :param water_level: water level used in remove instrument response. The
        default value in obspy is 60.
    :type water_level: float
    :param filter_flag: flag for filter the seismogram
    :type filter_flag: bool
    :param pre_filt: list of tuple of 4 corner frequency for filter,
        in ascending order(unit: Hz)
    :type pre_filt: list, tuple or numpy.array
    :param starttime: starttime of cutting
    :type starttime: obspy.UTCDateTime
    :param endtime: endtime of cutting
    :type endtime: obspy.UTCDateTime
    :param resample_flag: flag for data resampling
    :type resample_flag: bool
    :param sampling_rate: resampling rate(unit: Hz)
    :type sampling_rate: float
    :param taper_type: taper type, options from obspy taper
    :type taper_type: str
    :param taper_percentage: percentage of taper
    :type taper_percentage: float
    :param rotate_flag: rotate flag. If true, both inv and event location
        information should be provided
    :param event_latitude: latitude of event, for rotation usage
    :type event_latitude: float
    :param event_longitude: longitude of event, for rotation usage
    :type event_longitude: float
    :param sanity_check: sanity check the inventory information when
        rotating.
    :type sanity_check: bool
    :return: processed stream
    """
    # check input data type
    if isinstance(st, Trace):
        st = Stream(traces=[st, ])
        _is_trace = True
    elif isinstance(st, Stream):
        _is_trace = False
    else:
        raise TypeError("Input seismogram should be either obspy.Stream "
                        "or obspy.Trace")

    # cut the stream out before processing to reduce computation
    if starttime is not None and endtime is not None:
        st = flex_cut_stream(st, starttime, endtime, dynamic_npts=10)

    if filter_flag or remove_response_flag:
        # detrend ,demean, taper
        st.detrend("linear")
        st.detrend("demean")
        st.taper(max_percentage=taper_percentage, type=taper_type)

    # remove response or filter
    if filter_flag:
        if pre_filt is None or len(pre_filt) != 4:
            raise ValueError("Filter band should be list or tuple with "
                             "length of 4")
        if not check_array_order(pre_filt, order="ascending"):
            raise ValueError("Input pre_filt must be in ascending order: %s"
                             % pre_filt)

    if remove_response_flag:
        # remove response
        if inventory is None:
            raise ValueError("Station information(inv) should be provided if"
                             "you want to remove instrument response")
        st.attach_response(inventory)
        if filter_flag:
            st.remove_response(output="DISP", pre_filt=pre_filt,
                               zero_mean=False, taper=False,
                               water_level=water_level)
        else:
            st.remove_response(output="DISP", zero_mean=False, taper=False)
    elif filter_flag:
        # Perform a frequency domain taper like during the response removal
        # just without an actual response...
        filter_stream(st, pre_filt)

    if filter_flag or remove_response_flag:
        # detrend, demean or taper
        st.detrend("linear")
        st.detrend("demean")
        st.taper(max_percentage=taper_percentage, type=taper_type)

    # resample
    if resample_flag:
        # interpolation
        if sampling_rate is None:
            raise ValueError("sampling rate should be provided if you set"
                             "resample_flag=True")

        if endtime is not None and starttime is not None:
            npts = int((endtime - starttime) * sampling_rate) + 1
            st = interpolate_stream(st, sampling_rate, starttime=starttime,
                                    npts=npts)
        else:
            # it doesn't matter if starttime is None or not, cause
            # obspy will handle this case
            st = interpolate_stream(st, sampling_rate, starttime=starttime)
    else:
        if starttime is not None and endtime is not None:
            # just cut
            st.trim(starttime, endtime)

    # rotate
    if rotate_flag:
        st = rotate_stream(st, event_latitude, event_longitude,
                           inventory=inventory, mode="ALL->RT",
                           sanity_check=sanity_check)

    # Convert to single precision to save space.
    for tr in st:
        tr.data = np.require(tr.data, dtype="float32")

    # transfer back to trace if input type is Trace
    if _is_trace:
        st = st[0]

    return st



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

This script writes specfem sources into the respective simulation directories.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""

from gcmt3d.asdf.utils import smart_read_yaml
from gcmt3d.asdf.utils import is_mpi_env
from gcmt3d.data.management.create_process_paths import create_process_path_obs
from gcmt3d.data.management.create_process_paths import create_process_path_syn
from gcmt3d.data.management.create_process_paths import create_window_path
import os

# Get logger to log progress
from gcmt3d import logger


def make_paths(cmt_filename, param_path):

    # Define parameter directory
    databaseparam_path = os.path.join(param_path,
                                      "Database/DatabaseParameters.yml")

    # Load Parameters
    DB_params = smart_read_yaml(databaseparam_path,
                                mpi_mode=is_mpi_env())

    logger.info("Creating processing path files for the observed data...")

    # Create Processing path files observed
    process_obs_dir = os.path.join(param_path, "ProcessObserved")
    create_process_path_obs(cmt_filename, process_obs_dir)

    logger.info("Creating processing path files for the synthetic data...")

    # Create Processing path files synthetics
    process_syn_dir = os.path.join(param_path, "ProcessSynthetic")

    create_process_path_syn(cmt_filename, process_syn_dir, DB_params["npar"])

    logger.info("Creating processing path files for windowing the data...")

    # Create Window Path Files:
    window_dir = os.path.join(param_path, "CreateWindows")
    create_window_path(cmt_filename, window_dir,
                       figure_mode=DB_params["figure_mode"], verbose=True)
